<?php

namespace App\Http\Controllers;

use App\Events\UserCreated;
use App\Events\UserDeleted;
use App\Exports\UsersExport;
use App\Http\Requests\ImportRequest;
use App\Http\Requests\StoreUserRequest;
use App\Imports\UsersImport;
use App\Jobs\NotifyImportUserReady;
use App\Models\Group;
use App\Models\PhoneBook;
use App\Models\Role;
use App\Models\SipConversation;
use App\Models\User;
use App\Models\UserAgentProfile;
use App\Models\UserSetting;
use App\Models\VitXiConversation;
use Exception;
use Illuminate\Broadcasting\BroadcastException;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\QueryException;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Http\Response;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Hash;
use Illuminate\Support\Facades\Log;
use Maatwebsite\Excel\Facades\Excel;
use Maatwebsite\Excel\Validators\ValidationException;
use Pusher\ApiErrorException;
use Pusher\PusherException;
use Symfony\Component\HttpFoundation\BinaryFileResponse;

class AdminUserController extends Controller
{
	public function index(Request $request): JsonResponse
    {
        $user = auth()->user();

        $searchTerm = $request->input('search', '');
        $perPage = $request->input('per_page', 15);
        $tenant = $request->input('tenant', null);
        $orderBy = $request->input('order_by', 'updated_at');
        $sortBy = $request->input('sort_by', 'desc');

        $usersQuery = User::with(['roles', 'agent_profile', 'groups', 'phoneBooks']);

        if ($user->hasPermissionTo( 'all-tenants' )) {
            if (!is_null($tenant)) {
                $usersQuery = $usersQuery->where('tenant_id', $tenant);
            }
        } else {
            $usersQuery = $usersQuery->where('tenant_id', $user->tenant_id)
             ->whereHas('roles', function (Builder $q) {
                 $q->where('slug', '!=', 'super-admin');
             });
        }
        if (!empty($searchTerm)) {
            $usersQuery->where(function ($query) use ($searchTerm) {
                $query->where('users.name', 'like', '%' . $searchTerm . '%')
                      ->orWhere('users.user_name', 'like', '%' . $searchTerm . '%')
                      ->orWhere('users.email', 'like', '%' . $searchTerm . '%')
                      ->orWhere('users.extension', 'like', '%' . $searchTerm . '%');
            });
        }

        $users = $usersQuery
            ->orderBy($orderBy, $sortBy)
            ->paginate($perPage);
        return response()->json($users);
	}

	public function show(User $user): JsonResponse
    {
	    if (!auth()->user()->hasPermissionTo( 'all-tenants' )) {
            return response()->json(
                $user->with('roles', 'agent_profile', 'groups', 'phoneBooks')->get()
            );
        }
        return response()->json(
            $user
                ->where('tenant_id', '=', auth()->user()->tenant_id)
                ->with('roles', 'agent_profile', 'groups', 'phoneBooks')->get()
        );
	}

	public function store(StoreUserRequest $request)
	{
		$input = $request->all();
        $validated = $request->validated();

        // Prevents saving unwanted data in case the user sends these parameters
        $input['extension_id'] = null;
        $input['extension'] = null;
        $input['device_name'] = null;

        if (isset($validated['device_id'])) {
            $isVitXiDevice = $this->isVitXiDevice($validated['device_id'], $validated['tenant_id']);

            if (!$isVitXiDevice->vitxi_client) {
                return response()->json([
                    'message' => __('device_could_not_be_verified')
                ], 403);
            }
            $validated['extension_id'] = $isVitXiDevice->device->extension_id;
            if (!is_null($validated['extension_id'])) {
                $validated['extension'] = $isVitXiDevice->device->extension;
            } else {
                $validated['extension'] = $isVitXiDevice->device->assigned_exten;
            }

            if (!is_string($validated['extension'])) {
                return response()->json([
                    'message' => __('could_not_set_extension_to_user_please_try_again')
                ], 422);
            }
            $validated['device_name'] = $isVitXiDevice->device->device_name;

        }

		$validated['password'] = Hash::make($validated['password']);

		try {
			$user = User::create($validated);
            $role = Role::where('id', $validated['role'])->first();
            $user->roles()->attach($role);
            $user->groups()->attach($validated['groups']);

            if (isset($input['agent_profile_id']) && !is_null($input['agent_profile_id'])) {
                $userAgentProfile = new UserAgentProfile([
                    'agent_profile_id' => $input['agent_profile_id']
                ]);
                $user->agent_profile()->save($userAgentProfile);
            }

            if( isset($input['phone_books'])) {
                foreach ($input['phone_books'] as $item) {
                    $phoneBook = new PhoneBook();
                    $phoneBook->phone_book_id = $item['id'];
                    $phoneBook->description = $item['description'];
                    $phoneBook->type = $item['source_type'];
                    $phoneBook->tenant_id = $item['tenant_id'];
                    $user->phoneBooks()->save($phoneBook);
                }
            }

            /**
             * Set default settings for the user
             */
            $hasCallCenterFeatures = $user->hasPermissionTo('call-center-features');
            foreach ($user->defaultSettings() as $setting) {
                if ($setting['type'] === 'call_center') {
                    // add default call center settings
                    if ($hasCallCenterFeatures) {
                        $user_setting = new UserSetting();
                        $user_setting->type = $setting['type'];
                        $user_setting->key = $setting['key'];
                        $user_setting->value = $setting['value'];
                        $user->settings()->save($user_setting);
                    }
                } else {
                    $user_setting = new UserSetting();
                    $user_setting->type = $setting['type'];
                    $user_setting->key = $setting['key'];
                    $user_setting->value = $setting['value'];
                    if ($setting['key'] === 'timezone') {
                        $timezone = auth()->user()->searchSetting($setting['type'], $setting['key']);
                        $user_setting->value = !is_null($timezone) ? $timezone->value : 'UTC';
                    }
                    $user->settings()->save($user_setting);
                }
            }

            $user->userStatus()->create([
                'message' => '',
                'status_id' => 1
            ]);

            try {
                $eventLog = new PbxApiController();
                $user_aut = auth()->user();

                $eventLog->saveEvents(
                    $user_aut->tenant_id,
                    'Users',
                    'add',
                    $request->only([
                        'name',
                        'user_name',
                        'email',
                        'phone_books',
                        'tenant_id',
                        'agent_profile_id']) + [
                        'extension' => $request->extension,
                        'groups' => $user->groups()->pluck('name'),
                        'role' => $user->roles()->pluck('name')->first()
                    ],
                    $user_aut->user_name,
                    $user_aut->id,
                    [
                    ],
                    $request->ip(),
                    $request->userAgent()
                );
            } catch (Exception $e){

            }


            return response()->json([
                'message' => __('user_was_created'),
                'user' => $user->load(
                    'roles',
                    'groups',
                    'phoneBooks',
                    'agent_profile',
                )
            ]);
        } catch (\Exception $e) {
            return response()->json([
                'message' => $e->getMessage()
            ], 400);
        }
	}

    public function update(StoreUserRequest $request, User $user)
    {
        try {
            $validated = $request->validated();
            $input = $request->all();

            if ($user->id === 1 && $input['role'] !== 1) {
                throw new Exception(__('this_action_is_not_allowed'));
            }

            // Prevents saving unwanted data in case the user sends these parameters
            $input['extension_id'] = null;
            $input['extension'] = null;
            $input['device_name'] = null;

            if ($request->has('device_id')) {
                /**
                 * If the device is sent, the validation is carried out with the VitalPBX API,
                 * otherwise the values related to the device are set to null.
                 */
                if (!is_null($input['device_id'])) {
                    if (((string)$input['device_id'] !== $user->device_id) || is_null($user->device_name)) {
                        $deviceValidation = $this->isVitXiDevice($input['device_id'], $input['tenant_id'], $user->id);
                        if (!$deviceValidation->vitxi_client) {
                            return response()->json([
                                'message' => __('device_could_not_be_verified')
                            ], 403);
                        }
                        $user->device_id = $input['device_id'];
                        $user->extension_id = $deviceValidation->device->extension_id;

                        $extension = null;
                        if (!is_null($user->extension_id)) {
                            $extension = $this->getDeviceExtension($user->extension_id, $user->tenant_id);
                        } else {
                            $extension = $deviceValidation->device->assigned_exten;
                        }

                        if (!is_string($extension)) {
                            return response()->json([
                                'message' => __('could_not_set_extension_to_user_please_try_again')
                            ], 422);
                        }
                        $user->extension = $extension;
                        $user->device_name = $deviceValidation->device->device_name;
                    }
                } else {
                    $user->device_id = null;
                    $user->extension_id = null;
                    $user->extension = null;
                    $user->device_name = null;
                }
            }

            $user->name = $validated['name'];
            $user->user_name = $validated['user_name'];
            $user->email = $validated['email'];
            $user->tenant_id = $validated['tenant_id'];

            $user->roles()->sync([$validated['role']]);

            // remove call center settings if you have them
            DB::table('user_settings')
                ->where('user_id', '=', $user->id)
                ->where('type', '=', 'call_center')
                ->delete();
            $hasCallCenterFeatures = $user->hasPermissionTo('call-center-features');
            // Reassign call center permissions
            if ($hasCallCenterFeatures) {
                foreach ($user->defaultSettings() as $setting) {
                    if ($setting['type'] === 'call_center') {
                        $user_setting = new UserSetting();
                        $user_setting->type = $setting['type'];
                        $user_setting->key = $setting['key'];
                        $user_setting->value = $setting['value'];
                        $user->settings()->save($user_setting);
                    }
                }
            }

            $user->groups()->sync($input['groups']);

            /* set agent profile */
            if (isset($input['agent_profile_id']) && !is_null($input['agent_profile_id'])) {
                if ($user->agent_profile()->exists()) {
                    $user->agent_profile()->update([
                        'agent_profile_id' => $input['agent_profile_id']
                    ]);
                } else {
                    $userAgentProfile = new UserAgentProfile([
                        'agent_profile_id' => $input['agent_profile_id']
                    ]);
                    $user->agent_profile()->save($userAgentProfile);
                }
            }

            if (!empty($input['password'])) {
                $user->password = Hash::make($input['password']);
            }
            $user->update();

            if( isset($input['phone_books'])) {
                $user->phoneBooks()->delete();
                try {
                    foreach ($input['phone_books'] as $item) {
                        $phoneBook = new PhoneBook();
                        $phoneBook->phone_book_id = isset($item['id']) ? $item['id'] : $item['phone_book_id'];
                        $phoneBook->description = $item['description'];
                        $phoneBook->type = isset($item['source_type']) ? $item['source_type'] : $item['type'];
                        $phoneBook->tenant_id = $item['tenant_id'];
                        $user->phoneBooks()->save($phoneBook);
                    }
                } catch (Exception $e) {
                    return response()->json(['error' => __('error_on_update_phonebooks')], 401);
                }
            }

            /**
             * If the updated user is the current one, return the data required for the session,
             * in this way the local data is updated on the client side.
             */

            try {
                $eventLog = new PbxApiController();
                $auth_user = auth()->user();
                $eventLog->saveEvents(
                    $auth_user->tenant_id,
                    'Users',
                    'update',
                    $request->only([
                        'name',
                        'user_name',
                        'email',
                        'phone_books',
                        'tenant_id',
                        'agent_profile_id']) + [
                        'extension' => $request->extension,
                        'groups' => $user->groups()->pluck('name'),
                        'role' => $user->roles()->pluck('name')->first()
                    ],
                    $auth_user->user_name,
                    $auth_user->id,
                    [
                    ],
                    $request->ip(),
                    $request->userAgent()
                );
            }catch (Exception $e){

            }



            if ($user->id === auth()->id()) {
                return response()->json([
                    'message' => __('user_was_updated'),
                    'user' => $user->load([
                            'status',
                            'groups:id',
                            'settings:id,user_id,type,key,value',
                            'roles:id,slug,name',
                            'roles.permissions:id,name,slug',
                            'agent_profile.profile:id,name,crm_url',
                            'phoneBooks'
                        ]
                    )
                ], 200);
            }

            return response()->json([
                'message' => __('user_was_updated'),
                'user' => $user->load(
                    'roles',
                    'permissions',
                    'groups',
                    'phoneBooks',
                    'agent_profile',
                )
            ]);

        } catch (\Exception $e) {
            return response()->json(['message' => $e->getMessage()], 403);
        }
    }

	/*
	 * Get available roles to assign to users
	 */
	public function getAvailableRoles (): JsonResponse
    {
        try {
            if (auth()->user()->hasRole('super-admin')) {
                return response()->json(
                    Role::with('permissions')->get()
                );
            }

            return response()->json(
                Role::whereNotIn('slug', ['super-admin'])
                    ->with('permissions')->get()
            );
        } catch (Exception $exception) {
            return response()->json([
                'message' => $exception->getMessage()
            ], 422);
        }
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function getAvailableAgentProfiles (Request $request) {
        return ( new AgentProfileController )->index($request);
    }

	public function getDeviceExtension ($extension_id, $tenant_id) {
        /**
         * Make a request to the VitalPBX API to get the extension data
         */
        try {
            $response = ( new PbxApiController )->getExtension($extension_id, $tenant_id);

            if ($response->getStatusCode() === 200) {
                // $extension = $response->getBody()->read(2048);
                $extension = $response->getBody();
                $extension = json_decode($extension);
                return $extension->data->extension;
            }
        } catch (\Exception $exception) {

        }

        return null;
    }

    /**
     * Check if the device has been assigned y check if the selected device is marked as "VitXi Client".
     *
     * @param $device_id
     *
     * @param null $tenant_id
     * @return object
     */

	public function isVitXiDevice ($device_id, $tenant_id = null): object
    {
	    $validate = (object) [
            'vitxi_client' => false,
            'device' => null,
        ];
        // Make a request to the VitalPBX API to get the device data
        $response = ( new PbxApiController )->getDevice($device_id, $tenant_id);
        if ($response->getStatusCode() === 200) {
            $device = $response->getBody()->read(1024);
            $device = json_decode($device);
            if ($device->data->vitxi_client) {
                $validate->vitxi_client = true;
                $validate->device = $device->data;
            }
        }
        return $validate;
    }

    /**
     * @param $csvFile
     * @return false|int|string
     */
    public function detectDelimiter($csvFile)
    {
        $delimiters = [';' => 0, ',' => 0, "\t" => 0, '|' => 0];

        $handle = fopen($csvFile, 'rb');
        $firstLine = fgets($handle);
        fclose($handle);
        foreach ($delimiters as $delimiter => &$count) {
            $count = count(str_getcsv($firstLine, $delimiter));
        }

        return array_search(max($delimiters), $delimiters, true);
    }

    /**
     * @param ImportRequest $request
     *
     * @return JsonResponse|null
     */
    public function import (ImportRequest $request): ?JsonResponse {
        try {
            $file = $request->file('file');
            $delimiter = $this->detectDelimiter($file);
            $tenant = request('tenant') ?: auth()->user()->tenant_id;
            $auth_user = auth()->user();

            Excel::import(new UsersImport($delimiter, $tenant, $auth_user), $file)->chain([
                new NotifyImportUserReady($auth_user)
            ]);
            return response()->json([
                'message' => __('users_import_queued')
            ]);
        } catch (ValidationException $e) {
            $failures = $e->failures();
            return response()->json([
                'failures' => $failures
            ], 422);
        } catch (QueryException $e) {
            $message = $e->getMessage();
            $errorCode = $e->errorInfo[1];
            if($errorCode == '1062'){
                $message = __('one_of_the_users_or_some_of_the_data_already_exists_check_the_data_entered_and_try_again');
            }
            return response()->json([
                'message' => $message
            ], 400);
        }
    }

    /**
     * @param null $tenant_id
     * @param string $delimiter
     *
     * @return BinaryFileResponse|Response
     */
    public function export ($tenant_id = null, string $delimiter = 'comma'): Response|BinaryFileResponse {
        if ($delimiter === 'comma') {
            $delimiter = ',';
        } if ($delimiter === 'semicolon') {
            $delimiter = ';';
        }
        if (is_null($tenant_id) || !auth()->user()->hasPermissionTo('all-tenants')) {
            $tenant_id = auth()->user()->tenant_id;
        }

        return (new UsersExport($delimiter))
            ->forTenantID($tenant_id)
            ->download('users.csv', \Maatwebsite\Excel\Excel::CSV);
    }

    public function destroy(User $user)
	{
        try {
            if (!$user) {
                throw new Exception(__('it_is_not_allowed_to_delete_this_user'));
            }
            $groups = $user->groups()->get();
            $tenant_id = $user->tenant_id;

            $this->deleteConversations($user);

            if(!auth()->user()->hasPermissionTo('all-tenants')) {
                if ($tenant_id !== auth()->user()->tenant_id) {
                    return response()->json(['error' => __('not_allowed')], 403);
                }
            }

            $role = $user->roles()->pluck('name');

            $user->delete();
            try {
                if (count($groups) > 0) {
                    broadcast(new UserDeleted($user->id, $tenant_id, $groups));
                }
            } catch (\Exception $exception) {
                // return response()->json(['message' => $exception->getMessage()], 422);
            }

            try {
                $eventLog = new PbxApiController();
                $request = request();
                $user_aut = auth()->user();

                $eventLog->saveEvents(
                    $user_aut->tenant_id,
                    'Users',
                    'delete',
                    [
                        'name' => $user->name,
                        'user_name' => $user->user_name,
                        'email' => $user->email,
                        'extension' => $user->extension,
                        'phonebooks_id' => $user->phonebooks_id,
                        'tenant_id' =>$user->tenant_id,
                        'agent_profile_id'=>$user->agent_profile_id,
                        'groups' => $groups->pluck('name'),
                        'role' => $role
                    ],
                    $user_aut->user_name,
                    $user_aut->id,
                    [
                    ],
                    $request->ip(),
                    $request->userAgent()
                );
            }catch (Exception $e){

            }

            return response()->json(['message' => __('user_was_deleted')], 200);
        } catch (\Exception $e) {
            return response()->json(['message' => $e->getMessage()], 404);
        }
	}

    public function deleteConversations (User $user) {
        $conversations = VitXiConversation::forUser($user->id)->with('participants')->get()->unique();
        if ($user->device_name) {
            $sip_conversations = SipConversation::forUser($user->device_name)->with('participants')->get()->unique();
            $conversations = collect($conversations)->merge($sip_conversations)->all();
        }

        foreach ($conversations as $conversation) {
            if ($conversation->type === 'vitxi') {
                if (count($conversation->participants) < 2) {
                    $messages = $conversation->messages()->get();
                    foreach ($messages as $message) {
                        if (!is_null($message->attachment)) {
                            $file = $message->attachment->filepath . '/' . $message->attachment->filename;
                            if (file_exists($file)) {
                                unlink($file);
                            }
                            $message->attachment()->delete();
                        }
                    }
                    $conversation->delete();
                }
            } else if ($conversation->type === 'sip') {
                $conversation->delete();
            }
        }
    }
}
