<?php

namespace App\Http\Controllers;

use App\Http\Requests\AgentProfileRequest;
use App\Models\AgentProfile;
use Exception;
use Facade\FlareClient\Http\Exceptions\NotFound;
use Illuminate\Auth\Access\AuthorizationException;
use Illuminate\Database\Eloquent\ModelNotFoundException;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\DB;
use Illuminate\Validation\ValidationException;


class AgentProfileController extends Controller
{
    public function index (Request $request): JsonResponse {
        try {
            $query = AgentProfile::with('pauses');
            $user = auth()->user();
            if (!$user->hasPermissionTo('all-tenants')) {
                $query->where('tenant_id', $user->tenant_id);
            } elseif ($request->has('tenant_id')) {
                $tenantId = $request->input('tenant_id');
                $query->where('tenant_id', $tenantId);
            }

            return response()->json(['profiles' => $query->get()]);
        } catch (Exception $e) {
            return response()->json([
                'message' => $e->getMessage()
            ], 422);
        }

    }

    public function show ($id): JsonResponse
    {
        try {
            $user = auth()->user();
            $query = AgentProfile::with('pauses');

            // If the user is not an admin, limit the query to the current tenant_id
            if (!$user->hasPermissionTo('all-tenants')) {
                $query->where('tenant_id', $user->tenant_id);
            }

            // Find the agent profile by ID
            $profile = $query->findOrFail($id);

            return response()->json(compact('profile'));
        } catch (ModelNotFoundException $e) {
            return response()->json([
                'message' => 'Agent Profile not found.'
            ], 404);
        } catch (Exception $exception) {
            return response()->json([
                'message' => $exception->getMessage()
            ], 422);
        }
    }

    /**
     * Store a newly created resource in storage.
     *
     * @param Request $request
     *
     * @return JsonResponse
     */
    public function store(AgentProfileRequest $request) {
        try {

            $validated = $request->validated();
            $user = auth()->user();

            $is_global = $request->input('is_global') ?? false;
            $tenant_id = null;

            if (!$is_global) {
                $tenant_id = $request->input('tenant_id') ?? $user->tenant_id;
            }
            $this->authorize('create', [
                AgentProfile::class,
                $tenant_id,
                $is_global,
            ]);

            $profile = new AgentProfile();

            $profile->name = $validated['name'];
            $profile->is_global = $is_global;
            $profile->tenant_id = $tenant_id;
            $profile->crm_url = $validated['crm_url'] ?? null;
            $profile->website_tab_name = $validated['website_tab_name'];
            $profile->website_tab_url = $validated['website_tab_url'];
            $profile->save();


            $tmp = array();
            foreach($validated['pauses'] as $k => $v) {
                $tmp[$k] = $v['slug'];
            }

            // Find duplicates in temporary array
            $tmp = array_unique($tmp);
            // Remove the duplicates from original array
            foreach($validated['pauses'] as $k => $v)
            {
                if (!array_key_exists($k, $tmp)) {
                    unset($validated['pauses'][$k]);
                }
            }

            foreach ($validated['pauses'] as $value) {
                $profile->pauses()->create([
                    'name' => $value['name'],
                    'slug' => $value['slug']
                ]);
            }

            try {
                $eventLog = new PbxApiController();

                $pauses_name = [];
                foreach ($request['pauses'] as $pause){
                    $pauses_name [] = $pause['name'];
                }

                $user_aut = auth()->user();
                $eventLog->saveEvents(
                    $user_aut->tenant_id,
                    'Agent Profile',
                    'add',
                    $request->except(['created_at','updated_at','pauses']) + [
                        'pauses' => $pauses_name
                    ],
                    $user_aut->user_name,
                    $user_aut->id,
                    [
                    ],
                    $request->ip(),
                    $request->userAgent()
                );
            } catch (Exception $e) {
                // nothing to do here
            }
            return response()->json([
                'message' => __('globals_actions_create_success'),
                'profile' => $profile->load('pauses')
            ]);
        } catch (Exception $exception) {
            return response()->json([
                'message' => $exception->getMessage()
            ], 422);
        }
    }

    /**
     * Update the specified resource in storage.
     *
     * @param AgentProfileRequest $request
     * @param $id
     *
     * @return JsonResponse
     */
    public function update(AgentProfileRequest $request, $id) {
        try {
            $validated = $request->validated();
            $user = auth()->user();

            $is_global = $request->input('is_global') ?? false;
            $tenant_id = null;

            if (!$is_global) {
                $tenant_id = $request->input('tenant_id') ?? $user->tenant_id;
            }

            $profile = AgentProfile::findOrFail($id);

            $this->authorize('update', [
                $profile,
                $tenant_id,
                $is_global,
            ]);

            $profile->name = $validated['name'];
            $profile->tenant_id = $tenant_id;
            $profile->is_global = $is_global;
            $profile->crm_url = $validated['crm_url'] ?? null;
            $profile->website_tab_name = $validated['website_tab_name'];
            $profile->website_tab_url = $validated['website_tab_url'];
            $profile->save();

            // Build temporary array for array_unique
            $tmp = array();
            foreach($validated['pauses'] as $k => $v) {
                $tmp[$k] = $v['slug'];
            }

            // Find duplicates in temporary array
            $tmp = array_unique($tmp);
            // Remove the duplicates from original array
            foreach($validated['pauses'] as $k => $v)
            {
                if (!array_key_exists($k, $tmp)) {
                    unset($validated['pauses'][$k]);
                }
            }
            if ($profile->id === 1) {
                // default pauses (4)
                $default_pauses = ['administrative', 'lunch', 'break', 'wc'];

                DB::table('agent_pauses')
                    ->where('agent_profile_id', '=', 1)
                    ->whereNotIn('slug', $default_pauses)
                    ->delete();

                foreach($validated['pauses'] as $k => $v)
                {
                    if (isset($v['id'])) {
                        if (in_array(($v['slug'] ?? null), $default_pauses, true)) {
                            unset($validated['pauses'][$k]);
                        }
                    }
                }

                $uniqueValidation =  (object) [
                    'already_exists' => false,
                    'key' => ''
                ];

                foreach ($validated['pauses'] as $key => $value) {
                    if (!in_array(($value['slug'] ?? null), $default_pauses, true)) {
                        $profile->pauses()->create([
                            'name' => $value['name'],
                            'slug' => $value['slug']
                        ]);
                    } else {
                        $uniqueValidation->already_exists = true;
                        $uniqueValidation->key = $key;
                        break;
                    }
                }
                if ($uniqueValidation->already_exists) {
                    throw ValidationException::withMessages(['pauses.'. $uniqueValidation->key .'.name' => __('the_name_has_already_been_taken')]);
                }
            } else {
                $profile->pauses()->delete();
                foreach ($validated['pauses'] as $key => $value) {
                    $profile->pauses()->create([
                        'name' => $value['name'],
                        'slug' => $value['slug']
                    ]);
                }
            }
            try {
                $eventLog = new PbxApiController();

                $pauses_name = [];
                foreach ($request['pauses'] as $pause){
                    $pauses_name [] = $pause['name'];
                }

                $eventLog->saveEvents(
                    $user->tenant_id,
                    'Agent Profile',
                    'update',
                    $request->except(['id','created_at', 'updated_at','pauses']) + [
                        'pauses' => $pauses_name
                    ],
                    $user->user_name,
                    $user->id,
                    [
                    ],
                    $request->ip(),
                    $request->userAgent()
                );
            } catch (Exception $e){
                // nothing to do here
            }

            return response()->json([
                'message' => __('globals_actions_update_success'),
                'profile' => $profile->load('pauses')
            ]);
        } catch (ModelNotFoundException | ValidationException | Exception $exception) {
            return response()->json([
                'message' => $exception->getMessage()
            ], 422);
        }
    }

    /**
     * Remove the specified resource from storage.
     *
     * @param $id
     * @return JsonResponse
     * @throws AuthorizationException
     */
    public function destroy($id): JsonResponse
    {
        $profile = AgentProfile::findOrFail($id);
        $pauses = $profile->pauses->pluck('name');


        $this->authorize('delete', $profile);
        $request = request();
        $profile->delete();


        try {
            $eventLog = new PbxApiController();
            $user = auth()->user();
            $eventLog->saveEvents(
                $user->tenant_id,
                'Agent Profile',
                'delete',
                [
                    'name' => $profile->name,
                    'tenant_id' => $user->tenant_id,
                    'is_global' => $profile->is_global,
                    'website_tab_name' => $profile->website_tab_name,
                    'website_tab_url' => $profile->website_tab_url,
                    'crm_url'=> $profile->crm_url,
                    'pauses' => $pauses
                ],
                $user->user_name,
                $user->id,
                [
                ],
                $request->ip(),
                $request->userAgent()
            );
        } catch (Exception $e) {
            // nothing to do here
        }

        return response()->json([
            'message' => __('globals_actions_delete_success')
        ]);
    }
}
