<?php

namespace App\Http\Controllers;

use App\Exports\ContactsExport;
use App\Http\Requests\ContactDeleteFormRequest;
use App\Http\Requests\ContactsFormRequest;
use App\Imports\ContactsImport;
use App\Jobs\NotifyImportContactReady;
use App\Models\AppSetting;
use App\Models\Contact;
use App\Models\ContactPhone;
use App\Models\User;
use Exception;
use Illuminate\Database\Eloquent\ModelNotFoundException;
use Illuminate\Database\QueryException;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\File;
use Maatwebsite\Excel\Facades\Excel;
use Maatwebsite\Excel\Validators\ValidationException;

class ContactsController extends Controller
{
    /**
     * Display a listing of the resource.
     *
     * @param Request $request
     * @param bool $fromWsApi
     * @return JsonResponse|array
     */

    public function index(Request $request, bool $fromWsApi = false): JsonResponse|array {
        $searchTerm = trim($request->input('search', ''));
        $filterBy = $request->input('filter_by', 'all');
        $phoneType = $request->input('phone_type', '');
        $page = $request->input('page', 1);
        $perPage = $request->input('per_page') ?? 15;
        $sortBy = $request->input('sort_by', 'updated_at');
        $sortDesc = $request->input('sort_desc', 'desc');

        $userId = auth()->id();

        $contacts = Contact::with([
            'phones:id,contact_id,type,prefix,country_code,phone,blf',
        ])
            ->select('id', 'first_name', 'last_name', 'company', 'job_title', 'type', 'email', 'favorite')
            ->where(function ($query) use ($searchTerm, $phoneType) {
                $query->where(function ($query) use ($searchTerm) {
                    $query->where('first_name', 'like', '%' . $searchTerm . '%')
                        ->orWhere('last_name', 'like', '%' . $searchTerm . '%')
                        ->orWhere(DB::raw("concat(first_name, ' ', last_name)"), 'LIKE', "%".$searchTerm."%")
                        ->orWhere('email', 'like', '%' . $searchTerm . '%')
                          ->orWhere('company', 'like', '%' . $searchTerm . '%');
                })
                ->orWhereHas('phones', function ($subQuery) use ($searchTerm, $phoneType) {
                    $subQuery->where('phone', 'like', '%' . $searchTerm . '%');
                });

            })->whereHas('user', function ($query) use ($userId) {
                 $query->where('id', $userId);
            });

        if (!empty($phoneType)) {
            $contacts->whereHas('phones', function ($subQuery) use ($phoneType) {
                $subQuery->where('type', 'like', $phoneType);
            });
        }

        if ($filterBy === 'default') {
            $contacts->where('favorite', false);
        } elseif ($filterBy === 'favorites') {
            $contacts->where('favorite', true);
        } elseif ($filterBy === 'blf') {
            $contacts->whereHas('blfPhones');
        }

        $contacts = $contacts
                    ->orderBy('first_name', 'asc')
                    ->orderBy('last_name', 'asc')
                    ->paginate($perPage, ['*'], 'page', $page);

        return [
            'current_page' => $contacts->currentPage(),
            'data' => $contacts->items(),
            'first_page_url' => $contacts->url(1),
            'from' => $contacts->firstItem(),
            'last_page' => $contacts->lastPage(),
            'per_page' => $contacts->perPage(),
            'to' => $contacts->lastItem(),
            'total' => $contacts->total(),
        ];
    }

    /**
     * Store a newly created resource in storage.
     *
     * @param ContactsFormRequest $request
     *
     * @return JsonResponse
     * @throws \JsonException
     */
    public function store(ContactsFormRequest $request)
    {
        $validated = $request->validated();

        $user = auth()->user();

        $phones = json_decode($validated['phones'], true, 512, JSON_THROW_ON_ERROR);
	    $phoneExist = [];

        $phonesWithBlf = [];
        foreach ($phones as $phone) {
            if (isset($phone['blf']) && $phone['blf'] === true) {
                $phonesWithBlf[] = $phone;
            }
            if ($user->getContactByPhone($phone)) {
                $phoneExist[] = $phone;
            }
        }

        if (count($phoneExist) > 0) {
            return response()->json([
                'message' => __('the_phone_number_already_exists_please_try_another_number'),
                'phones' => $phoneExist
            ], 422);
        }
        if((count($phonesWithBlf) > 0) && $user->hasReachedBlfLimit()) {
            return response()->json([
                'message' => 'Exceeded BLF contact limit of phones with BLF enabled'
            ], 400);
        }

	    try {

		    $contact = new Contact($validated);

		    if ($request->hasFile('avatar')) {
                $contact->addMediaFromRequest('avatar')->toMediaCollection('avatar');
		    }

		    auth()->user()->contacts()->save($contact);

		    foreach ($phones as $phone) {
		    	$contact_phone = new ContactPhone($phone);
		        $contact->phones()->save($contact_phone);
		    }

		 return response()->json([
            'message' => __('the_contact_was_created_successfully'),
            'contact' => $contact->load('phones')
         ]);
	    } catch (Exception $e) {
		    return response()->json([
                'message' => $e->getMessage()
            ], $e->getCode());
	    }
    }

    /**
     * Display the specified resource.
     *
     * @param $id
     *
     * @return JsonResponse
     */
    public function show($id)
    {
    	$contact = auth()->user()->contacts()
                    ->with('phones')->where('id', $id)
            ->first();

    	if ($contact) {
	        return response()->json(compact('contact'));
	    } else {
    		return response()->json(['message' => __('the_contact_does_not_exist')], 404);
	    }
    }

    public function getContactByPhone ($phone, $fromWsConnection = false) {
        $contact = auth()->user()->getContactByPhone($phone);

        if ($fromWsConnection) {
            return $contact;
        }
        if ($contact) {
            return response()->json(compact('contact'));
        } else {
            return response()->json(['message' => __('the_contact_does_not_exist')], 404);
        }
    }

    /**
     * Update the specified resource in storage.
     *
     * @param ContactsFormRequest $request
     * @param $id
     *
     * @return JsonResponse
     */
    public function update(ContactsFormRequest $request, $id): JsonResponse
    {
    	try {
            $validated = $request->validated();
            $user = auth()->user();

		    $contact = $user->contacts()->where('id', $id)->firstOrFail();
		    $contact->first_name = $validated['first_name'];
		    $contact->last_name = $validated['last_name'];
		    $contact->company = $validated['company'] ?? null;
		    $contact->job_title = $validated['job_title'] ?? null;
		    $contact->type = $validated['type'];
		    $contact->email = $validated['email'] ?? null;
		    $contact->favorite = $validated['favorite'] ?? false;


            if ($request->hasFile('avatar')) {
                $contact->addMediaFromRequest('avatar')->toMediaCollection('avatar');
            } else {
                if ($contact->hasMedia('avatar')) {
                    $contact->clearMediaCollection('avatar');
                }
            }

		    $contact->save();

            // Get all the contact's current phones from the database
            $existingPhones = $contact->phones->keyBy('id');

            // Get the phones sent from the frontend
            $phonesToUpdate = json_decode($validated['phones'], true, 512, JSON_THROW_ON_ERROR);

            // Get the current ContactPhone number with blf set to true associated with the user
            $userBlfPhonesCount = $user->contactsWithBlf()->count();

            $blfLimit = 50;
            $limitSetting = AppSetting::where('type', 'blf', 'limit')->first();
            if (!is_null($limitSetting)) {
                $blfLimit = (int)$limitSetting->value;
            }

            foreach ($phonesToUpdate as $phoneData) {
                // If the phone already exists in the database, update the relevant fields
                if (isset($phoneData['id'], $existingPhones[$phoneData['id']])) {
                    $existingPhone = $existingPhones[$phoneData['id']];
                    // Check if the blf field has changed from false to true
                    if (!$existingPhone->blf && $phoneData['blf']) {
                        // Validate the ContactPhone limit with blf to true for the user
                        if ($userBlfPhonesCount >= $blfLimit) {
                            return response()->json([
                                'message' => 'Exceeded BLF contact limit of phones with BLF enabled'
                            ], 422);
                        }
                        $userBlfPhonesCount++;
                    }

                    $existingPhone->update([
                        'type' => $phoneData['type'],
                        'phone' => $phoneData['phone'],
                        'prefix' => $phoneData['prefix'],
                        'country_code' => $phoneData['country_code'],
                        'blf' => $phoneData['blf'],
                    ]);
                } else {
                    // Validate if the phone is already assigned to another contact
                    $existingContact = $user->getContactByPhone($phoneData['phone']);
                    if ($existingContact && $existingContact->id !== $contact->id) {
                        // The phone is already assigned to another contact
                        return response()->json([
                            'message' => 'The phone ' . $phoneData['phone'] . ' is already assigned to another contact [' . $existingContact->full_name .']'
                        ], 422);
                    }
                    // Validate the ContactPhone limit with blf to true for the user
                    if ($phoneData['blf'] && $userBlfPhonesCount >= $blfLimit) {
                        return response()->json([
                            'message' => 'Exceeded BLF contact limit of phones with BLF enabled'
                        ], 400);
                    }
                    // If the phone does not exist in the database, create a new one and save it associated with the contact
                    $phone = new ContactPhone([
                        'type' => $phoneData['type'],
                        'phone' => $phoneData['phone'],
                        'prefix' => $phoneData['prefix'],
                        'country_code' => $phoneData['country_code'],
                        'blf' => $phoneData['blf'],
                    ]);
                    $contact->phones()->save($phone);

                    // Increment the counter if the phone has blf to true
                    if ($phoneData['blf']) {
                        $userBlfPhonesCount++;
                    }
                }
            }

            // Delete phones that existed but are no longer present on the frontend
            $existingPhoneIds = $contact->phones->pluck('id')->filter()->toArray();
            $phonesToDeleteIds = array_diff($existingPhoneIds, array_column($phonesToUpdate, 'id'));
            ContactPhone::whereIn('id', $phonesToDeleteIds)->delete();

            return response()->json([
                'message' => __('the_contact_has_been_updated_successfully'),
                'contact' => $contact->load('phones')
            ]);

	    } catch (ModelNotFoundException $e) {
		    return response()->json(['message' => $e->getMessage()], 404);
	    }  catch (Exception $e) {
            return response()->json([
                'message' => $e->getMessage()
            ], 422);
        }
    }

    public function toggleFavorite($id): JsonResponse
    {
        $contact = Contact::where('user_id', auth()->id())->findOrFail($id);
        $contact->favorite = !$contact->favorite;
        $contact->save();

        return response()->json([
            'message' => $contact->favorite ? __('contact_added_successfully_to_favorites') : __('contact_removed_successfully_from_favorites'),
            'favorite' => $contact->favorite
        ]);
    }

    /**
     * Get all the contacts mark as favorites.
     *
     * @param $userId
     *
     * @return JsonResponse
     */
    public function getFavorites($userId = null): JsonResponse
    {
        $favorites = Contact::where('user_id', $userId ?? auth()->id())
            ->where('favorite', true)
            ->with('contact.phones')
            ->get();
        if ($favorites->isEmpty()) {
            return response()->json(['message' => __('no_favorites')], 422);
        }
        return response()->json([
            'favorites' => $favorites
        ]);
    }
    /**
     * @param $csvFile
     * @return false|int|string
     */
    public function detectDelimiter($csvFile)
    {
        $delimiters = [';' => 0, ',' => 0, "\t" => 0, '|' => 0];

        $handle = fopen($csvFile, 'rb');
        $firstLine = fgets($handle);
        fclose($handle);
        foreach ($delimiters as $delimiter => &$count) {
            $count = count(str_getcsv($firstLine, $delimiter));
        }

        return array_search(max($delimiters), $delimiters, true);
    }

    public function import (Request $request) {
        try {
            $file = $request->file('file');
            $existingAction = $request->input('existing');
            $delimiter = $this->detectDelimiter($file);
            $auth_user = auth()->user();
            Excel::import(new ContactsImport($delimiter, $existingAction, $auth_user), $file)->chain([
                new NotifyImportContactReady($auth_user)
            ]);
            return response()->json([
                'message' => __('contacts_import_queued')
            ]);
        } catch (ValidationException $e) {
            $failures = $e->failures();
            return response()->json([
                'failures' => $failures
            ], 422);
        } catch (QueryException $e) {
            $message = $e->getMessage();
            $errorCode = $e->errorInfo[1];
            if($errorCode == '1062'){
                $message = __('one_of_the_users_or_some_of_the_data_already_exists_check_the_data_entered_and_try_again');
            }
            return response()->json([
                'message' => $message
            ], 422);
        }
    }

    public function export ($delimiter = 'comma') {
        if ($delimiter === 'comma') {
            $delimiter = ',';
        } if ($delimiter === 'semicolon') {
            $delimiter = ';';
        }
        return (new ContactsExport($delimiter))->forUser(auth()->id())
            ->download('contacts.csv', \Maatwebsite\Excel\Excel::CSV);
    }

	/**
	 * Remove the specified resource from storage.
	 *
	 * @param $id
	 *
	 * @return JsonResponse
	 */
    public function destroy(ContactDeleteFormRequest $request)
    {
        try {
            $user = auth()->user();

            $contacts = $user->contacts()->whereIn('id', $request->ids)->get();

            if ($contacts->isEmpty()) {
                return response()->json(['message' => __('contact_not_found')], 404);
            }
            $contacts->each(function ($contact) {
                $contact->delete();
            });
            return response()->json(['message' => __('the_contact_has_been_deleted_successfully')]);

        } catch (Exception $exception) {
            return response()->json(['error' => $exception->getMessage()], 422);
        }
    }
}
