<?php
namespace App\Http\Controllers;
use App\Http\Requests\ConversationCreateForm;
use App\Http\Requests\SendMessageToConversationRequest;
use App\Models\SipConversation;
use App\Models\SipMessage;
use App\Models\User;
use App\Models\VitXiConversation;
use App\Models\VitXiMessage;
use App\Notifications\newVitXiMessage;
use Exception;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\ModelNotFoundException;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Http\Response;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\File;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\Facades\Notification;
use Illuminate\Support\Facades\Storage;
use Ramsey\Uuid\Uuid;
use Spatie\MediaLibrary\MediaCollections\Models\Media;
use stdClass;


class ConversationsController extends Controller
{

    /**
     * Show all the message conversations to the user.
     *
     * @param Request $request
     * @param bool $fromWsApi
     * @return mixed
     */
    public function index(Request $request, bool $fromWsApi = false): mixed
    {
        $conversations = [];
        $type = $request->input('type', 'all');

        try {
            switch ($type) {
                case 'all':
                    $conversations = $this->getVitXiConversations();
                    if (auth()->user()->device_name) {
                        $sipConversations = $this->getSipConversations();
                        $conversations = collect($conversations)->merge($sipConversations)->all();
                    }
                    break;
                case 'vitxi':
                    $conversations = $this->getVitXiConversations();
                    break;
                case 'sip':
                    if (auth()->user()->device_name) {
                        $conversations = $this->getSipConversations();
                    }
                    break;
            }
            return !$fromWsApi ? response()->json($conversations) : $conversations;
        } catch (Exception $e) {
            $exception = ['message' => $e->getMessage()];
            return !$fromWsApi ? response()->json($exception, 422) : $exception;
        }
    }

    public function getVitXiConversations() {
        return auth()->user()->vitxiConversations()
                     ->with([
                           'participants.user:id,name,extension,email,presence',
                           'participants.user.status',
                           'latestMessage.media',
                           ])
                            ->where(function ($query) {
                            $query->whereNull('deleted_by')
                                ->orWhere('deleted_by', '<>', auth()->id());
                        })
                     ->get();
    }
    public function getSipConversations() {
        return auth()->user()->sipConversations()
                     ->with([
                           'participants.contact.phones',
                           'latestMessage'
                     ])
                     ->get();
    }

    public function getByUuid (Request $request) {
        try {
            $conversation = new stdClass();
            if ($request->has('uuid')) {
                $uuid = $request->input('uuid');
                $userId = auth()->id();
                $conversation = VitXiConversation::where('uuid', $uuid)
                    ->where(function ($query) {
                        $query->whereNull('deleted_by')->orWhere('deleted_by', '<>', auth()->id());
                    })
                    ->with(['participants' => function ($query) use ($userId) {
                        $query->where('user_id', '!=', $userId)
                            ->with([
                                'user:id,name,extension,email,presence',
                                'user.status'
                            ]);
                    }])->first();

                if (!$conversation && auth()->user()->device_name) {
                    $conversation = SipConversation::findByUuid($uuid)
                        ->with('participants.contact.phones')
                        ->first();
                }
            }
            if (is_null($conversation) || count(get_object_vars($conversation)) == 0) {
                return response()->json(['message' => __('not_found')], 404);
            }

            return response()->json($conversation);
        } catch (Exception $exception) {
            return response()->json(['message' => $exception->getMessage()], 422);
        }
    }

    public function search ($type, $peer, bool $fromWsApi = false) {
        $conversation = null;

        switch($type) {
            case 'vitxi':
                $participantsId = [auth()->id(), $peer];
                $conversation = VitXiConversation::between($participantsId)->first();
                break;
            case 'sip':
                $participantsPhone = [auth()->user()->device_name, $peer];
                $conversation = SipConversation::between($participantsPhone)->first();
                break;
        }
        return $fromWsApi ? $conversation : response()->json($conversation);
    }

    /**
     * Shows a message conversation.
     *
     * @param string $type
     * @param $id
     * @param int $page
     * @return mixed
     */
    public function getMessages ($type = 'vitxi', $id, $page = 1)
    {
        try {
            switch ($type) {
                case 'vitxi':
                    $conversation = auth()->user()->vitxiConversations()
                        ->where(function (Builder $q) use ($id) {
                            $q->where('vitxi_conversations'.'.id', $id)
                                ->where(function ($query) {
                                    $query->whereNull('deleted_by')
                                        ->orWhere('deleted_by', '<>', auth()->id());
                                });
                        })->firstOrFail();
                    $conversation->markAsRead(auth()->id());
                    break;
                case 'sip':
                    $conversation = auth()->user()->sipConversations()
                        ->where(function (Builder $q) use ($id) {
                            $q->where('sip_conversations'.'.id', $id);
                        })->firstOrFail();
                    $conversation->markAsRead(auth()->user()->device_name);
                    break;
                default:
                    $conversation = auth()->user()->vitxiConversations()
                        ->where(function (Builder $q) use ($id) {
                            $q->where('vitxi_conversations'.'.id', $id);
                        })->firstOrFail();
            }

            $messages = $conversation->visibleMessages($page);
            return compact('messages');  // Response from Web Socket
        } catch (Exception $e) {
            return ['message' => $e->getMessage()];  // Response from Web Socket
        }
    }

    public function getMedia ($conversationID, $page = 1) {
        try {
            $conversation = auth()->user()->vitxiConversations()
                ->where(function (Builder $q) use ($conversationID) {
                    $q->where('vitxi_conversations'.'.id', $conversationID)
                        ->where(function ($query) {
                            $query->whereNull('deleted_by')
                                ->orWhere('deleted_by', '<>', auth()->id());
                        });
                })->firstOrFail();

            return $conversation->visibleMedia($page);  // Response from Web Socket
        } catch (Exception $e) {
            return ['message' => $e->getMessage()];  // Response from Web Socket
        }
        // Removed response from Http.
    }

    public function getAudio( $fileName ) {
        $filePath = storage_path('/uploads/audios/' . $fileName);

        try {
            return response()->download($filePath, $fileName, ['audio/mpeg']);
        } catch (Exception $exception) {
            return response()->json(['message' => $exception->getMessage()], 402);
        }
    }

    /**
     * Stores a new conversation.
     *
     * @param ConversationCreateForm $request
     *
     * @return JsonResponse
     */
    public function store(ConversationCreateForm $request): JsonResponse
    {
        try {
            $validated = $request->validated();
            $recipients = $validated['recipients'];
            $channel = $validated['channel'];

            if ($channel === 'sip') {
                // Validate if the recipient has chat permissions
                if ($validated['originator'] === 'local') {
                    $user = auth()->user();
                    if (!$user->hasPermissionTo('chat')) {
                        return response()->json(['message' => __('not_allowed')]
                       , 401);
                    }
                } elseif ($validated['originator'] === 'remote') {
                    $user = User::where('id', $recipients[0])->first();
                    if ($user && !$user->hasPermissionTo('chat')) {
                        return response()->json(['message' => __('recipient_no_chat_permission')]
                        , 400);
                    }
                }
            } else {
                $user = User::where('id', $recipients[0])->first();
                if ($user && !$user->hasPermissionTo('chat')) {
                    return response()->json(['message' => __('recipient_no_chat_permission')]
                        , 400);
                }
            }

            $conversation = match($channel) {
                'vitxi' => $this->storeVitXi($recipients),
                'sip' =>  $this->storeSip($recipients, $validated['originator']),
            };
            return response()->json($conversation);
        } catch (Exception $exception) {
            return response()->json(['error' => $exception->getMessage()], 422);
        }
    }

    public function storeVitXi(array $recipients) {
        $userId = auth()->id();
        $recipients[] = $userId;
        $participants = $recipients;
        $conversation = VitXiConversation::between($participants)->first();
        if (is_null($conversation)) {
            $conversation = VitXiConversation::create( [
                'type' => 'vitxi',
                'uuid'    => Uuid::uuid4()
            ] );
            $conversation->addParticipant($participants);
        } else if ($conversation->deleted_by === $userId) {
            $conversation->update([
                'deleted_by' => null
            ]);
        }
        return $conversation->load([
            'participants.user:id,name,extension,email,presence',
            'participants.user.status',
            'latestMessage.media',
        ]);
    }

    public function storeSip(array $recipients, string $originator) {
        $deviceName = auth()->user()->device_name;
        if (is_null($deviceName)) {
            throw new \RuntimeException(__('can_not_send_sip_messages'));
        }

        if ($originator === 'local') {
            $userExtension = DB::table('users')
                ->whereIn('extension', $recipients)
                ->where('tenant_id', auth()->user()->tenant_id)
                ->first();
            if ($userExtension) {
                throw new \RuntimeException(__('sip_conversation'));
            }
        }

        $recipients[] = $deviceName;
        $participants = $recipients;
        $conversation = SipConversation::between($participants)->first();
        if (is_null($conversation)) {
            $conversation = SipConversation::create( [
                'type' => 'sip',
                'uuid'    => Uuid::uuid4()
            ] );
            $conversation->addParticipant($participants);
        }
        return $conversation->load([
            'participants.contact.phones',
            'latestMessage'
        ]);
    }

    /**
     * Adds a new message to a current conversation.
     *
     * @param SendMessageToConversationRequest $request
     * @param $channel
     * @param $id
     *
     * @return JsonResponse
     */
    public function update(SendMessageToConversationRequest $request, $channel, $id)
    {
        $validated = $request->validated();
        try {
            $response = match ($channel) {
                'vitxi' => $this->sendVitXiMessage($id, $validated, $request->file('media')),
                'sip' => $this->sendSipMessage($id, $validated),
            };
            return response()->json($response);
        } catch (ModelNotFoundException $e) {
            return response()->json(['message' => $e->getMessage()], 404);
        } catch (Exception $e) {
            $statusCode = ($e->getCode() >= 100 && $e->getCode() < 600) ? $e->getCode() : 400;

            return response()->json([
                'message' => $e->getMessage()
            ], 400);
        }
    }

    public function sendVitXiMessage ($id, $input, $attachment = null) {
        $userId = auth()->id();
        $conversation = VitXiConversation::where('id', $id)
            ->where(function ($query) {
                $query->whereNull('deleted_by')->orWhere('deleted_by', '<>', auth()->id());
            })->first();
        $type = $input['type'] ?? 'text';
        $body = $input['body'] ?? null;
        $reply_to = $input['reply_to'] ?? null;
        $attributes = $input['attributes'] ?? null;

        //validate if the recipient has chat permissions
        foreach ($conversation->participants as $participant) {
            if ($participant->user_id !== $userId) {
                $recipient = User::where('id', $participant->user_id)->first();
                if (!$recipient->hasPermissionTo('chat')){
                    throw new \RuntimeException(__('recipient_no_chat_permission'));
                }
            }
        }

        $message = VitXiMessage::create([
            'conversation_id' => $id,
            'user_id' => $userId,
            'type' => $type,
            'body' => $body,
            'reply_to' => $reply_to
        ]);

        $conversation->markAsRead($userId);
        if (!is_null($attachment)) {
            $message
                ->addMediaFromRequest('media')
                ->withCustomProperties([
                    'attributes' => $attributes,
                ])
                ->toMediaCollection('attachment');
        }
        foreach ($conversation->participants as $participant) {
            if ($participant->user_id !== $userId) {
                $peer = User::where('id', $participant->user_id)->first();
                Notification::sendNow($peer, new newVitXiMessage($message, $conversation));

                // $peer->notify(new newVitXiMessage($message, $conversation));
            }
        }
        return [
            'conversation' => $conversation->load([
                'participants.user:id,name,extension,email,presence',
                'participants.user.status',
                'latestMessage.media',
            ]),
            'message' => $message->load('reference.media', 'media')
        ];
    }


    public function sendSipMessage ($id, $input) {
        $deviceName = auth()->user()->device_name;
        if (is_null($deviceName)) {
            throw new \RuntimeException(__('can_not_receive_send_sip_messages'));
        }
        $conversation = SipConversation::findById($id)->firstOrFail();
        $body = $input['body'];
        $phone = $input['originator'] === 'local' ? $deviceName : $input['phone'];
        $message = SipMessage::create([
            'conversation_id' => $id,
            'phone' => $phone,
            'type' => 'text',
            'body' => $body,
        ]);
        if ($input['originator'] === 'local') {
        	$conversation->markAsRead($deviceName);
        }
        return [
            'conversation' => $conversation->load([
                'participants.contact.phones',
                'latestMessage'
            ]),
            'message' => $message
        ];
    }

    public function unreadQueryThreads($assignment)
    {
        $unreadThreads = [];
        if ($threads = VitXiConversation::where('subject', 'LIKE', '%query-' . $assignment->id . '%')->get()) {
            foreach ($threads as $thread) {
                if ($thread->isUnread(Authorizer::getResourceOwnerId())) {
                    array_push($unreadThreads, $thread->subject);
                }
            }
        }
        return $unreadThreads;
    }

    public function leaveConversation($id)
    {
        try {
            $conversation = VitXiConversation::findOrFail($id);
            $conversation->getParticipantFromUser(Auth::id());
        } catch (ModelNotFoundException $e) {
            Flash::error(__('the_requested_conversation_does_not_exist'));
            return redirect('conversations');
        }
        $participant = $conversation->getParticipantFromUser(Auth::id());
        $user = $participant->user;
        $conversation->messages()->where('user_id', '=', $user->id)->delete();
        $participant->delete();
        if ($conversation->participants()->count() == 0) {
            $conversation->delete();
        }
        Flash::success(__('you_have_left_the_conversation'));
        return redirect('conversations');
    }

    // Mark Messages as read from WebSocket and Http
    public function markAsRead($type = 'vitxi', $id, $fromWsApi = false)
    {
        try {
            switch ($type) {
                case 'vitxi':
                    $conversation = VitXiConversation::findOrFail($id);
                    $conversation->markAsRead(auth()->id());
                    break;

                case 'sip':
                    $conversation = SipConversation::findOrFail($id);
                    $conversation->markAsRead(auth()->user()->device_name);
                    break;
            }
            if (!$fromWsApi){
                return response()->json(['message' => __('success')]); // Response Http
            }
            return ['message' => __('success')]; // Response Web Socket
        }
        catch(Exception $e){
            if (!$fromWsApi){
                return response()->json(['message' => $e->getMessage()], 422); // Response Http
            }
            return ['message' => $e->getMessage()];  // Response from Web Socket
        }
    }

    /** Delete a conversation
     * @param $type
     * @param $id
     * @return JsonResponse
     */
    public function deleteConversation($type, $id, $fromWsApi = false)
    {
        try {
            if ($type === 'vitxi') {
                $conversation = auth()->user()->vitxiConversations()->findOrFail($id);
            } else {
                $conversation = auth()->user()->sipConversations()->findOrFail($id);
                $conversation->delete();
                if (!$fromWsApi){
                    return response()->json(['message' => __('the_conversation_was_deleted')]); // Response Http
                }
                return ['message' => __('the_conversation_was_deleted')]; // Response Web Socket
            }

            if (is_null($conversation->deleted_by) && count($conversation->participants) > 1) {
                $conversation->deleted_by = auth()->id();
                $conversation->save();
            } else {
                if ($conversation->deleted_by !== auth()->id() || count($conversation->participants) < 2) {
                    if ($conversation->type === 'vitxi') {
                        $messages = $conversation->messages()->get();
                        foreach ($messages as $message) {
                            if (!is_null($message->attachment)) {
                                $file = $message->attachment->filepath . '/' . $message->attachment->filename;
                                if (file_exists($file)) {
                                    unlink($file);
                                }
                                $message->attachment()->delete();
                            }
                        }
                    }
                    $conversation->delete();
                }
            }
        } catch (Exception $e) {
            if (!$fromWsApi){
                return response()->json(['message' => $e->getMessage()], 404); // Response Http
            }
            return ['message' => $e->getMessage()]; // Response Web Socket
        }
        if (!$fromWsApi){
            return response()->json(['message' => __('the_conversation_was_deleted')]); // Response Http
        }
        return ['message' => __('the_conversation_was_deleted')]; // Response Web Socket
    }

    /**
     * Delete a message
     *
     * @param $conversation_type
     * @param $conversation_id
     * @param $message_id
     *
     * @return JsonResponse
     */
    public function deleteMessage($conversation_type, $conversation_id, $message_id)
    {
        if ($conversation_type !== 'vitxi' && $conversation_type !== 'sip'){
            return ['message' => __('invalid_conversation_type')];  // Response from Web Socket
        }   // Check if the message type is a valid one

        try {
            if ($conversation_type === 'vitxi') {
                $conversation = auth()->user()->vitxiConversations()->findOrFail($conversation_id);
                $message = $conversation->messages()->findOrFail($message_id);
            } else {
                $conversation = auth()->user()->sipConversations()->findOrFail($conversation_id);
                $message = $conversation->messages()->findOrFail($message_id);
            }
        } catch (ModelNotFoundException $e) {
            return ['message' => $e->getMessage()];
        }

        $conversation_id = $message->conversation_id;

        if ($conversation_type === 'sip') {
            $message->delete();
        } else {
            if (is_null($message->deleted_by)) {
                $message->deleted_by = auth()->id();
                $message->save();
            } else {
                if ($message->deleted_by !== auth()->id()) {
                    if (!is_null($message->attachment)) {
                        $file = $message->attachment->filepath . '/' . $message->attachment->filename;
                        if (file_exists($file)) {
                            unlink($file);
                        }

                        $message->attachment()->delete();
                    }
                    $message->delete();
                } else {
                    return ['message' => __('message_not_found')];
                }
            }
        }

        if ($conversation->messages()->count() === 0) {
            $conversation->delete();
            return ['message' => __('the_conversation_was_deleted')];
        }

        return ['message' => __('message_deleted')];  // Response from Web Socket
    }

}
