<?php

namespace App\Http\Controllers;

use App\Http\Middleware\SetLocale;
use Carbon\Carbon;
use Exception;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Log;
use JsonException;
use Ratchet\ConnectionInterface;
use Ratchet\RFC6455\Messaging\MessageInterface;
use Ratchet\WebSocket\MessageComponentInterface;
use BeyondCode\LaravelWebSockets\WebSockets\Exceptions\UnknownAppKey;
use BeyondCode\LaravelWebSockets\Apps\App;
use Illuminate\Support\Facades\Auth;
use Tymon\JWTAuth\Exceptions\JWTException;
use Tymon\JWTAuth\Exceptions\TokenExpiredException;
use Tymon\JWTAuth\Exceptions\TokenInvalidException;
use Tymon\JWTAuth\Facades\JWTAuth;
use BeyondCode\LaravelWebSockets\WebSockets\Channels\PresenceChannel;
class CustomAppWebSocketHandler implements MessageComponentInterface
{

    public function onOpen(ConnectionInterface $conn)
    {
        $appKey = 'vitxi4561aq';
        if (is_null(App::findByKey($appKey))) {
            throw new UnknownAppKey($appKey);
        }

        $conn->app = new \stdClass();
        $conn->app->id = 'vitxi4561aq';

        $socketId = sprintf('%d.%d', random_int(1, 1000000000), random_int(1, 1000000000));

        $conn->socketId = $socketId;
    }

    public function onClose(ConnectionInterface $conn)
    {
        // TODO: Implement onClose() method.
    }

    public function onError(ConnectionInterface $conn, Exception $e): void
    {
        Log::error('Connection error: ' . $e);
        $conn->close();
    }

    /**
     * @throws JsonException
     */
    public function onMessage(ConnectionInterface $conn, MessageInterface $msg): void
    {
        $user = null;
        try {
            // Decode the message
            $msg = json_decode($msg, false, 512, JSON_THROW_ON_ERROR);

            // Attempt to set and validate token
            JWTAuth::setToken($msg->token ?? null);
            $user = JWTAuth::toUser($msg->token);

            // If no user is found, throw an exception
            if (empty($user)) {
                throw new Exception('User not found');
            }

            // Set locale using the user's preferences
            (new SetLocale())->handlews($user->locale);

        } catch (TokenExpiredException | TokenInvalidException | JWTException | Exception $e) {
            // Invalid token response
            $user = null;
            $this->invalidToken($conn);
            return;
        }

        switch ($msg->event ?? '') {
            case 'ping':
                $conn->send(json_encode([
                    'event' => 'pong',
                    'requestId' => $msg->requestId,
                    'resource' => null,
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Get my info.
            case 'me':
                $conn->send(json_encode([
                    'event' => 'me',
                    'requestId' => $msg->requestId,
                    'resource' => (new UserController)->me(true, $user->id),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Get all groups.
            case 'get_groups':
                $conn->send(json_encode([
                    'event' => 'get_groups',
                    'requestId' => $msg->requestId,
                    'resource' => (new UserController)->getGroups(true),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Get all users from a group.
            case 'get_users_from_group':
                $conn->send(json_encode([
                    'event' => 'get_users_from_group',
                    'requestId' => $msg->requestId,
                    'resource' => (new UserController)->getGroupUsers(
                        new Request([
                            'search' => $msg->payload->search ?? '',
                            'page' => $msg->payload->page ?? null,
                            'per_page' => $msg->payload->per_page ?? null,
                        ]),
                        $msg->payload->group_id,
                        true,
                    ),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Get all users.
            case 'get_users':
                $conn->send(json_encode([
                    'event' => 'get_users',
                    'requestId' => $msg->requestId,
                    'resource' => (new UserController)->index(
                        new Request([
                            'search' => $msg->payload->search ?? '',
                            'page' => $msg->payload->page ?? null,
                            'per_page' => $msg->payload->per_page ?? null,
                            'filter_by' => $msg->payload->filter_by ?? null,
                        ]),
                        true
                    ),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Get user by extension.
            case 'find_user_by_extension':
                $conn->send(json_encode([
                    'event' => 'find_user_by_extension',
                    'requestId' => $msg->requestId,
                    'resource' => UserController::getUserByExtension($msg->payload->extension, $user->tenant_id),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Get contacts.
            case 'get_contacts':
                $conn->send(json_encode([
                    'event' => 'get_contacts',
                    'requestId' => $msg->requestId,
                    'resource' => (new ContactsController)->index(new Request([
                        'search' => $msg->payload->search ?? '',
                        'page' => $msg->payload->page ?? null,
                        'per_page' => $msg->payload->per_page ?? null,
                        'filter_by' => $msg->payload->filter_by ?? null,
                        'phone_type' => $msg->payload->phone_type ?? '',
                    ]), true),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Get favorites.
            case 'get_favorites':
                $conn->send(json_encode([
                    'event' => 'get_favorites',
                    'requestId' => $msg->requestId,
                    'resource' => (new ContactsController)->getFavorites($user->id),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Get contact by phone.
            case 'find_contact_by_phone':
                $conn->send(json_encode([
                    'event' => 'find_contact_by_phone',
                    'requestId' => $msg->requestId,
                    'resource' => (new ContactsController)->getContactByPhone($msg->payload->phone, true),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            // Get calls.
            case 'get_calls':
                $response = ( new PbxApiController() )->getCalls(new Request([
                    'peer' => $msg->payload->peer ?? null,
                    'start_date' => $msg->payload->start_date ?? null,
                    'end_date' => $msg->payload->end_date ?? null,
                    'disposition' => $msg->payload->disposition ?? null,
                    'recordings' => $msg->payload->recordings ?? null,
                    'search' => $msg->payload->search ?? null,
                    'offset' => $msg->payload->offset ?? null,
                    'row_count' => $msg->payload->row_count ?? null,
                ]));
                $conn->send( json_encode( [
                    'event'    => 'get_calls',
                    'requestId' => $msg->requestId,
                    'resource' => $response->getStatusCode() === 200 ? $response->getData() : null,
                    'message' => $response->getStatusCode() !== 200 ? $response->getData() : null,
                    'error'    => $response->getStatusCode() !== 200,
                ], JSON_THROW_ON_ERROR ) );
                break;
            case 'get_conversations':
                $conn->send(json_encode([
                    'event' => 'get_conversations',
                    'requestId' => $msg->requestId,
                    'resource' => (new ConversationsController())->index(
                        new Request([
                            'type' => $msg->payload->type,
                        ]),
                        true,
                    ),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            // search for conversation
            case 'search_conversation':
                $conn->send(json_encode([
                    'event' => 'search_conversation',
                    'requestId' => $msg->requestId,
                    'resource' => (new ConversationsController())->search(
                        $msg->payload->type,
                        $msg->payload->peer,
                        true,
                    ),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Get message.
            case 'get_messages':
                $conn->send(json_encode([
                    'event' => 'get_messages',
                    'requestId' => $msg->requestId,
                    'resource' => (new ConversationsController())->getMessages($msg->payload->conversation_type, $msg->payload->conversation_id, $msg->payload->page),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Delete message.
            case 'delete_message':
                $conn->send(json_encode([
                    'event' => 'delete_message',
                    'requestId' => $msg->requestId,
                    'resource' => (new ConversationsController())->deleteMessage($msg->payload->conversation_type, $msg->payload->conversation_id, $msg->payload->message_id),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Delete Conversation.
            case 'delete_conversation':
                $conn->send(json_encode([
                    'event' => 'delete_conversation',
                    'requestId' => $msg->requestId,
                    'resource' => (new ConversationsController())->deleteConversation($msg->payload->conversation_type, $msg->payload->conversation_id, true),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            // Get media.
            case 'get_media':
                $conn->send(json_encode([
                    'event' => 'get_media',
                    'requestId' => $msg->requestId,
                    'resource' => (new ConversationsController())->getMedia($msg->payload->conversation_id, $msg->payload->page),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            // todo
            /* case 'send_message':
                $data = array(
                    'foo' => 'bar'
                );
                $request = new Request();
                $request->setMethod('POST');
                $request->request->add(
                    $msg->payload
                );
                $message = (new ConversationsController())->update($request, $msg->payload->conversation_id);
                $conn->send(json_encode([
                    'event' => 'send_message',
                    'resource' => $message,
                    'error' => false
                ]));
                break; */
            //Mark as read.
            case 'mark_as_read':
                $conn->send(json_encode([
                    'event' => 'mark_as_read',
                    'requestId' => $msg->requestId,
                    'resource' => (new ConversationsController())->markAsRead($msg->payload->conversation_type, $msg->payload->conversation_id, true),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Get phonebooks.
            case 'get_phonebooks':
                $conn->send(json_encode([
                    'event' => 'get_phonebooks',
                    'requestId' => $msg->requestId,
                    'resource' => (new PhoneBookController())->index(true),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //Get phonebooks contacts.
            case 'get_phonebooks_contacts':
                $conn->send(json_encode([
                    'event' => 'get_phonebooks_contacts',
                    'requestId' => $msg->requestId,
                    'resource' => (new PbxApiController())->getPhonebookContacts($msg->payload->phonebook_id, true),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
            //EOL
            case 'get_queues_status':
                $conn->send(json_encode([
                    'event' => 'get_queues_status',
                    'requestId' => $msg->requestId,
                    'resource' => (new PbxApiController())->getDeviceQueuesStatusWs($user->device_id),
                    'error' => false
                ], JSON_THROW_ON_ERROR));
                break;
        }

    }

    /**
     * @throws JsonException
     */
    public function invalidToken (ConnectionInterface $conn): void
    {
        $conn->send(
            json_encode([
                'message' => __('invalid_or_expired_user_credentials'),
                'error' => true,
                'code' => 401
            ], JSON_THROW_ON_ERROR)
        );
        $conn->close();
    }

}
