<?php

namespace App\Http\Controllers;

use App\Http\Requests\GroupRequest;
use App\Models\Group;
use Exception;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Http\Response;

class GroupsController extends Controller
{
    public function index(Request $request): JsonResponse {
        $user = auth()->user();
        $isSuperAdmin = $user->hasPermissionTo('all-tenants');
        $tenantId = null;

        if ($isSuperAdmin) {
            // If the user is a super-admin, check if a tenant_id parameter is in the request
            $tenantId = $request->input('tenant_id', null);

            if ($tenantId) {
                // If tenant_id is provided in the request, filter by it
                $groups = Group::where('tenant_id', $tenantId)
                    ->withCount('users')
                    ->get();
            } else {
                // If no tenant_id is provided, return all groups
                $groups = Group::all()
                    ->loadCount('users');
            }
        } else {
            // If the user is not a super-admin, filter by the tenant_id of the authenticated user
            $tenantId = $user->tenant_id;
            $groups = Group::where('tenant_id', $tenantId )
                ->withCount('users')
                ->get();
        }
        // If there are no groups for the specified tenant, create a new one
        if ( !is_null( $tenantId ) && $groups->isEmpty() ) {
            // Create the default group for main tenant
            Group::factory()
                ->count(1)
                ->create([
                    'name' => 'Default group',
                    'description' => 'Generic group created by default.',
                    'tenant_id' => $tenantId
                ]);
            $groups = Group::where( 'tenant_id', $tenantId )
                           ->withCount( 'users' )
                           ->get();
        }
        return response()->json($groups);

    }

    public function store (GroupRequest $request): JsonResponse
    {
        $validated = $request->validated();
        $this->authorize('create', [
            Group::class,
            $validated['tenant_id'],
        ]);
        $group = Group::create($validated);

        try {
            $eventLog = new PbxApiController();
            $user_auth = auth()->user();
            $eventLog->saveEvents(
                $user_auth->tenant_id,
                'Groups',
                'add',
                [
                    'name' => $group->name,
                    'description' => $group->description,
                    'tenant_id' => $group->tenant_id
                ],
                $user_auth->user_name,
                $user_auth->id,
                [
                ],
                $request->ip(),
                $request->userAgent()
            );
        } catch (Exception $e) {
            // nothing to do here
        }
        return response()->json([
            'message' => __('globals_actions_create_success'),
            'group' => $group->loadCount('users')
        ], 201);
    }

    public function show(Group $group): JsonResponse
    {
        $this->authorize('view', $group);
        return response()->json($group->loadCount('users'));
    }

    // create a method to get the users ids that belong to a group
    public function getMembersIds($id): JsonResponse
    {
        $group = Group::findOrFail($id);
        $this->authorize('view', $group);
        $membersIds = $group->users->pluck('id')->toArray();
        return response()->json($membersIds);
    }

    public function syncMembers($id, Request $request): JsonResponse
    {
        // validate the request to ensure that the members_ids array is present
        $request->validate(['members_ids' => 'required|array']);

        $group = Group::findOrFail($id);
        $this->authorize('update', $group);

        $membersIds = $request->input('members_ids', []);
        $group->users()->sync($membersIds);
        $group->touch();
        return response()->json([
            'message' => 'Members updated successfully',
            'group' => $group->loadCount('users'),
        ]);
    }

    public function update (GroupRequest $request, Group $group): JsonResponse
    {
        $validated = $request->validated();
        $this->authorize('update', $group);
        $group->update($validated);

        $eventLog = new PbxApiController();
        $user_auth = auth()->user();
        $users = $group->users()->get();
        $users_data = [];
        foreach ($users as $user){
            $users_data []= [
                'name' => $user->name,
                'extension' => $user->extension
            ];
        }

        try {
            $eventLog->saveEvents(
                $request->tenant_id,
                'Groups',
                'update',
                [
                    'name' => $group->name,
                    'description' => $group->description,
                    "agents"=>$users_data

                ],
                $user_auth->user_name,
                $user_auth->id,
                [
                ],
                $request->ip(),
                $request->userAgent()
            );
        } catch (Exception $e){
            // nothing to do here
        }
        return response()->json([
            'message' => __('globals_actions_update_success'),
            'group' => $group->loadCount('users')
        ]);
    }

    public function destroy (Group $group): Response
    {
        $this->authorize('delete', $group);

        $users = $group->users()->get();
        $users_data = [];
        foreach ($users as $user){
            $users_data []= [
                'name' => $user->name,
                'extension' => $user->extension
            ];
        }

       $group->delete();

        try {
            $eventLog = new PbxApiController();
            $user_auth = auth()->user();
            $request = request();

            $eventLog->saveEvents(
                $user_auth->tenant_id,
                'Groups',
                'delete',
                [
                    'name' => $group->name,
                    'description' => $group->description,
                ],
                $user_auth->user_name,
                $user_auth->id,
                [
                ],
                $request->ip(),
                $request->userAgent()
            );
        } catch (Exception $e){
            // nothing to do here
        }

        return response()->noContent();
    }
}
