<?php

namespace App\Http\Controllers;

use App\Http\Requests\UpdateEmailSettingsRequest;
use App\Utils\Environment;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Artisan;
use Illuminate\Support\Facades\Config;
use Illuminate\Support\Facades\Mail;
use Illuminate\Support\Facades\Validator;
use Symfony\Component\Mailer\Exception\TransportExceptionInterface;

class MailSettingsController extends Controller
{
    private $email_settings;

    public function __construct () {

        $this->email_settings = [
            'type' => config('mail.driver'),
            'host' => config('mail.host'),
            'port' => config('mail.port'),
            'username' => config('mail.username'),
            'password' => config('mail.password'),
            'encryption' => config('mail.encryption'),
            'from_address' => config('mail.from.address'),
            'from_name' => config('mail.from.name'),
            'bcc_mail' => config('mail.bcc_mail')
        ];
    }

    public function show () {
        return response()->json($this->email_settings);
    }

    public function update (UpdateEmailSettingsRequest $request) {
        $validated = $request->validated();

        try {

            $from_name = $validated['from_name'];

            $new_values = [
                'MAIL_DRIVER' => $validated['type'],
                'MAIL_HOST' => $validated['host'] ?? '',
                'MAIL_PORT' => $validated['port'] ?? '',
                'MAIL_USERNAME' => $validated['username'] ?? '',
                'MAIL_PASSWORD' => $validated['password'] ?? '',
                'MAIL_ENCRYPTION' => $validated['encryption'] ?? '',
                'MAIL_FROM_ADDRESS' => $validated['from_address'],
                'MAIL_FROM_NAME' => "\"$from_name\"",
                'MAIL_BCC_MAIL' => $validated['bcc_mail'],
            ];

            $updated = Environment::updateSettings($new_values);

            if (!$updated) {
                return response()->json(['message' => __('could_not_update')], 422);
            }

            Artisan::call('cache:clear');
            Artisan::call('config:clear');

            $this->email_settings = $validated;

            $eventLogs = new PbxApiController();
            $user_aut = auth()->user();


            $eventLogs->saveEvents(
                $user_aut->tenant_id,
                'Mail Settings',
                "update",
                $request->input(),
                $user_aut->user_name,
                $user_aut->id,
                [
                ],
                $request->ip(),
                $request->userAgent()
            );

            return response()->json($this->email_settings);

        } catch (\Exception $exception) {
            // nothing
            return response()->json($exception);
        }
    }

    public function sendTestEmail (UpdateEmailSettingsRequest $request) {
        $validated = $request->validated();

        $rules = [
            'email_to_send_test' => 'required|email'
        ];

        $validator = Validator::make($request->all(), $rules);

        if ($validator->fails()) {
            return response()->json($validator->errors(), 422);
        }
        try {
            Config::set('mail.driver', $validated['type']);
            Config::set('mail.host', $validated['host'] ?? '');
            Config::set('mail.port', $validated['port'] ?? '');
            Config::set('mail.username', $validated['username'] ?? '');
            Config::set('mail.password', $validated['password'] ?? '');
            Config::set('mail.encryption', $validated['encryption'] ?? '');
            Config::set('mail.from.address', $validated['from_address']);
            Config::set('mail.from.name', $validated['from_name']);
            Config::set('mail.from.name', $validated['from_name']);

            Mail::send('Email.test', [], function ($message) use ($request) {
                $message->to($request->input('email_to_send_test'))
                    ->subject('WebRTCApp - Test email');
            });
            return response()->json([
                'message' => __('the_email_was_sent_successfully')
            ]);
        } catch (TransportExceptionInterface $exception) {
            return response()->json([
                'message' => $exception->getMessage()
            ], 422);
        } catch (\Exception $exception) {
            return response()->json([
                'message' => $exception->getMessage()
            ], 422);
        }
    }

}
