<?php

namespace App\Http\Controllers;

use App\Http\Requests\DeviceRequest;
use App\Models\Contact;
use App\Models\User;
use Carbon\Carbon;
use Exception;
use Firebase\JWT\JWT;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ConnectException;
use GuzzleHttp\Exception\RequestException;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Support\Arr;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Http;
use Illuminate\Support\Facades\Log;
use UnexpectedValueException;

class PbxApiController extends Controller
{
	private string $base;
	private string $sub;
	private Client $client;
	private mixed $key;

	public function __construct()
	{
        $this->middleware('auth:api', ['except' => []]);
        $protocol = 'https://';
        $host = config('app.pbx.host');
        $pbxHttpsPort = config('app.pbx.api_port');

        $this->base = $protocol . $host . ':'. $pbxHttpsPort;
        $version = 'v2';
        $this->sub = '/api/' . $version . '/';
        $this->key = config('app.pbx.app_key');
    	$this->client = new Client();
	}

    public function getUri () {
        return $this->base . $this->sub;
    }

    public function getKey () {
        return $this->key;
    }


	private function requestHeader ($tenant_id = null, $contentType = 'application/json') {
	    return [
            'Accept'     => 'application/json',
            'Content-Type' => $contentType,
            'Access-Control-Allow-Headers' => 'Authorization',
            'app-key' => $this->key,
            'tenant' => $tenant_id ?: auth()->user()->tenant_id
        ];
    }

    public function getEventLogs(Request $request) {
        try {
            $secretKeyFile = '/etc/vitalpbx/vitalpbx-maint.conf';

            if (!file_exists($secretKeyFile)) {
                return false;
            }
            // Payload data you want to include in the JWT
            $jwtPayload = array(
                "app" => "VitXi"
            );


            $secretKey = file_get_contents($secretKeyFile);

            // Build JWT
            $jwt = JWT::encode($jwtPayload, $secretKey, 'HS256');

            $endpoint = 'https://'.config('app.pbx.host').'/events-logs/v1/api/events';

            $timezone = auth()->user()->searchSetting('general', 'timezone');
            $timezone = ( !empty($timezone->value) ) ? $timezone->value : 'UTC';

            $startDate = $request->input('start_date');
            $endDate = $request->input('end_date');
            if (!is_null($startDate) && !is_null($endDate)) {
                $startDate = Carbon::createFromFormat("Y-m-d H:i:s", $startDate, $timezone)->timestamp;
                $endDate = Carbon::createFromFormat("Y-m-d H:i:s", $endDate, $timezone)->timestamp;
            }

            $user = auth()->user();
            $isSuperAdmin = $user->hasRole('super-admin');

            $params = [
              'app' => 'vitxi',
              'search' => trim($request->input('search','')),
              'tenant_id' => $request->input('tenant_id', !$isSuperAdmin ? auth()->user()->tenant_id : null),
              'module' => $request->input('module', null),
              'page' => $request->input('page', 1),
              'per_page' => $request->input('per_page', 10),
              'sort_by' => $request->input('sort_by', 'created_at'),
              'start_date' => $startDate,
              'end_date' => $endDate,
            ];

            $response = Http::withToken($jwt)
                            ->withHeaders([
                                'Accept'     => 'application/json',
                                'Content-Type' => 'application/json',
                            ])->withoutVerifying()
                            ->get($endpoint, $params);

            if (!$response->successful()){
                Log::warning(__('failed_to_log_event' . $response->body()));
            }
            // traverse the response data and convert the date (it's in timestamp) field to the format Y-m-d H:i:s
            $data = json_decode($response->body(), true);
            $data['data'] = Arr::map($data['data'], function ($item, $key) use ($timezone) {
                $item['date'] = Carbon::createFromTimestamp($item['date'])->setTimezone($timezone)->toDateTimeString();
                return $item;
            });
            return $data;
        } catch (Exception $e) {
            // nothing to do here
            return response()->json(['message' => $e->getMessage()]);
        }
    }

    public function saveEvents(
        $tenant_id,
        $module,
        $event,
        $payload,
        $username,
        $user_id,
        $metadata,
        $client_ip,
        $device_info,
        $app = 'vitxi',
    ): bool {
        try {
            $secretKeyFile = '/etc/vitalpbx/vitalpbx-maint.conf';

            if (!file_exists($secretKeyFile)) {
                return false;
            }

            $secretKey = file_get_contents($secretKeyFile);


            // Payload data you want to include in the JWT
            $jwtPayload = array(
                "app" => $app
            );


            // Build JWT
            $jwt = JWT::encode($jwtPayload, $secretKey, 'HS256');

            $endpoint = 'https://'.config('app.pbx.host').'/events-logs/v1/api/events';

            $response = Http::withToken($jwt)
                ->withHeaders([
                    'Accept'     => 'application/json',
                    'Content-Type' => 'application/json',
                ])->withoutVerifying()
                ->post($endpoint,
                    [
                        'app' => $app,
                        'tenant_id' => $tenant_id,
                        'module' => $module,
                        'event' => $event,
                        'payload' => $payload,
                        'username' => $username,
                        'user_id' => $user_id,
                        'metadata' => $metadata,
                        'client_ip' => $client_ip,
                        'device_info' =>$device_info
                    ]
                );

            if (!$response->successful()){
                Log::warning(__('failed_to_log_event' . $response->body()));
            }
        } catch (Exception $e) {
            // nothing to do here
        }

        return true;

    }
    public function makeRequest ($method, $path, $tenant_id = null, $isWsRequest = false, $query = null)
    {
        try {
            if (is_null($tenant_id)) {
                $tenant_id = request('tenant') ?: auth()->user()->tenant_id;
            }
            $uri = $this->base . $this->sub . $path;

            $options = [
                'headers' => [
                    'Accept'     => 'application/json',
                    'Content-Type' => 'application/json',
                    'Access-Control-Allow-Headers' => 'Authorization',
                    'app-key' => $this->key,
                    'tenant' => $tenant_id
                ],
                'verify' => false
            ];

            if (!is_null($query)) {
                $options['query'] = $query;
            }

            $res = $this->client->request($method, $uri, $options);
            $body = $res->getBody();
            return $isWsRequest ? json_decode($body, true, 512, JSON_THROW_ON_ERROR) : $body;
        } catch (RequestException $e) {
            return response()->json(['error' => $e->getMessage()], $e->getCode());
        }
    }

    public function getCurrentPlan () {
	    try {
            $response = $this->client->get($this->base . $this->sub . 'core/current_plan', [
                'headers' => [
                    'Accept'     => 'application/json',
                    'Content-Type' => 'application/json',
                    'Access-Control-Allow-Headers' => 'Authorization',
                    'app-key' => $this->key
                ],
                'verify' => false
            ]);
            if ($response->getStatusCode() === 200) {
                return json_decode($response->getBody())->data;
            }
        } catch (RequestException | ConnectException | ClientException $exception) {
            return response()->json(['error' => $exception->getMessage()], 503);
        }
    }

    public function getTenants(Request $request, $fromCli = false) {
        $id = '';
        if(!$fromCli && !$request->user()->hasPermissionTo('all-tenants')) {
            $id = auth()->user()->tenant_id;
        }

	    $resource = isset($id) && !empty($id) ? 'tenants/' . $id : 'tenants';
	    try {
		    $request = $this->client->get($this->base . $this->sub . $resource, [
			    'headers' => [
				    'Accept'     => 'application/json',
				    'Content-Type' => 'application/json',
				    'Access-Control-Allow-Headers' => 'Authorization',
				    'app-key' => $this->key,
			    ],
			    'verify' => false
		    ]);
            return $request->getBody();
	    } catch (RequestException $exception) {
		    return response()->json(['error' => $exception->getMessage()], 422);
	    }

    }

    public function getExtensions(Request $request) {

	    if(!$request->user()->hasPermissionTo('all-tenants')) {
            if ($request->input('tenant') != auth()->user()->tenant_id) {
                return response()->json(['error' => 'not allowed'], 403);
            }
        }

        $id = $request->input('id');
        $tenant = $request->input('tenant');
		$resource = isset($id) ? 'extensions/' . $id : 'extensions';

		try {
            $response = $this->client->get($this->base . $this->sub . $resource, [
			    'headers' => [
				    'Accept'     => 'application/json',
				    'Content-Type' => 'application/json',
				    'Access-Control-Allow-Headers' => 'Authorization',
				    'app-key' => $this->key,
				    'tenant' => $tenant
			    ],
			    'verify' => false
		    ]);
		    return $response->getBody();
		} catch (RequestException  $exception) {
			return response()->json(['error' => $exception->getMessage()], $exception->getCode());
		}
    }

    public function getExtension ($extension_id, $tenant_id) {

        $resource = 'extensions/' . $extension_id;
        try {
            /**
             * Returns the entire response in case the function is called from another controller.
             */
            return $this->client->get( $this->base . $this->sub . $resource, [
                'headers' => [
                    'Accept'     => 'application/json',
                    'Content-Type' => 'application/json',
                    'Access-Control-Allow-Headers' => 'Authorization',
                    'app-key' => $this->key,
                    'tenant' => $tenant_id
                ],
                'verify' => false
            ]);
        } catch (RequestException  $exception) {
            return response()->json(['error' => $exception->getMessage()], $exception->getCode());
        }
    }

    public function getDevices(Request $request) {
        $tenant = $request->input('tenant') ?: auth()->user()->tenant_id;
        return $this->makeRequest('GET', 'devices/vitxi', $tenant);
    }

    public function getDevice($device_id = null, $tenant_id = null) {
        if (is_null(auth()->user()->device_id) && is_null($device_id)) {
            return response()->json([
                'error' => __('you_do_not_have_a_device_assigned')
            ], 422);
        }
        $resource = 'devices/' . (is_null($device_id) ? auth()->user()->device_id : $device_id);
        try {
            $response = $this->client->get($this->base . $this->sub . $resource, [
                'headers' => [
                    'Accept'     => 'application/json',
                    'Content-Type' => 'application/json',
                    'Access-Control-Allow-Headers' => 'Authorization',
                    'app-key' => $this->key,
                    'tenant' => $tenant_id ?: auth()->user()->tenant_id
                ],
                'verify' => false
            ]);

            /**
             * Returns the entire response in case the function is called from another controller.
             */
            if (!is_null($device_id)) {
                return $response;
            }

            /**
             * Returns the payload when the function is called from the client (frontend).
             */
            return $response->getBody();

        } catch (RequestException  $exception) {
            return response()->json(['error' => $exception->getMessage()], $exception->getCode());
        }
    }

    /**
     * Gets the queues assigned to the user's device
     * @param DeviceRequest $request
     * @return JsonResponse
     */
    public function getDynamicQueues (DeviceRequest $request) {
        try {
            $user = auth()->user();
            $resource = 'devices/' . $user->device_id . '/queues/dynamic';

            $response = $this->client->get($this->base . $this->sub . $resource, [
                'headers' => $this->requestHeader(),
                'verify' => false
            ]);

            if ($response->getStatusCode() !== 200) {
                return response()->json(['message' => 'error on get queues'], 422);
            }

            $response = json_decode($response->getBody(), true);

            return response()->json($response['data']);

        } catch (Exception $exception) {
            return response()->json(['error' => $exception->getMessage()], $exception->getCode());
        }
    }

    public function loginToQueues (DeviceRequest $request) {
        try {
            $user = auth()->user();
            $resource = 'devices/' . $user->device_id . '/queues-login';
            $queues = $request->has('queues') ? $request->input('queues') : 'all';

            if (!$user->hasPermissionTo('queue-selection')) {
                $queues = 'all';
            }

            $response = $this->client->post($this->base . $this->sub . $resource, [
                'headers' => $this->requestHeader(null, 'application/x-www-form-urlencoded'),
                'verify' => false,
                'form_params' => [
                    'queues' => $queues
                ]
            ]);

            if ($response->getStatusCode() !== 200) {
                return response()->json(['message' => __('could_not_login')], 422);
            }

            $response = json_decode($response->getBody(), true);
            return response()->json($response);


        } catch (Exception $exception) {
            return response()->json(['error' => $exception->getMessage()], $exception->getCode());
        }
    }

    public function getDeviceQueuesStatus (DeviceRequest $request) {
        try {
            $user = auth()->user();
            $resource = 'devices/' . $user->device_id . '/queues/status';

            /* if (!$user->hasPermissionTo('queue-selection')) {
                return response()->json(['message' => 'you do not have access to this action'], 422);
            } */

            $response = $this->client->get($this->base . $this->sub . $resource, [
                'headers' => $this->requestHeader(),
                'verify' => false
            ]);

            if ($response->getStatusCode() !== 200) {
                return response()->json(['message' => __('the_action_could_not_be_performed')], 422);
            }

            $response = json_decode($response->getBody(), true);
            return response()->json($response);


        } catch (Exception $exception) {
            return response()->json(['error' => $exception->getMessage()], $exception->getCode());
        }
    }

    public function getDeviceQueuesStatusWs ($device_id) {
        try {
            $resource = 'devices/' . $device_id . '/queues/status';

            $response = $this->client->get($this->base . $this->sub . $resource, [
                'headers' => $this->requestHeader(),
                'verify' => false
            ]);

            if ($response->getStatusCode() !== 200) {
                return ['message' => __('the_action_could_not_be_performed')];
            }

            return json_decode($response->getBody(), true, 512, JSON_THROW_ON_ERROR);


        } catch (Exception $exception) {
            return ['message' => $exception->getMessage()];
        }
    }

    public function pause (DeviceRequest $request) {
        try {
            $user = auth()->user();
            $resource = 'devices/' . $user->device_id . '/queues-pause';

            $queues = $request->has('queues') ? $request->input('queues') : 'all';
            $reason = $request->has('reason') ? $request->input('reason') : 'break';

            $response = $this->client->post($this->base . $this->sub . $resource, [
                'headers' => $this->requestHeader(null, 'application/x-www-form-urlencoded'),
                'verify' => false,
                'form_params' => [
                    'queues' => $queues,
                    'reason' => $reason
                ]
            ]);

            if ($response->getStatusCode() !== 200) {
                return response()->json(['message' => __('the_action_could_not_be_performed')], 422);
            }

            $response = json_decode($response->getBody(), true);

            return response()->json($response);

        } catch (Exception $exception) {
            return response()->json(['error' => $exception->getMessage()], $exception->getCode());
        }
    }

    public function unpause (DeviceRequest $request) {
        try {
            $user = auth()->user();
            $resource = 'devices/' . $user->device_id . '/queues-unpause';

            $queues = $request->has('queues') ? $request->input('queues') : 'all';

            $response = $this->client->post($this->base . $this->sub . $resource, [
                'headers' => $this->requestHeader(null, 'application/x-www-form-urlencoded'),
                'verify' => false,
                'form_params' => [
                    'queues' => $queues
                ]
            ]);

            if ($response->getStatusCode() !== 200) {
                return response()->json(['message' => __('the_action_could_not_be_performed')], 422);
            }

            $response = json_decode($response->getBody(), true);

            return response()->json($response);

        } catch (Exception $exception) {
            return response()->json(['error' => $exception->getMessage()], $exception->getCode());
        }
    }

    public function logoutToQueues (DeviceRequest $request) {
        try {
            $user = auth()->user();
            $resource = 'devices/' . $user->device_id . '/queues-logout';

            if (!$user->hasPermissionTo('queue-selection')) {
                $queues = 'all';
            } else {
                $queues = $request->has('queues') ? $request->input('queues') : 'all';
            }

            $response = $this->client->post($this->base . $this->sub . $resource, [
                'headers' => $this->requestHeader(null, 'application/x-www-form-urlencoded'),
                'verify' => false,
                'form_params' => [
                    'queues' => $queues
                ]
            ]);

            if ($response->getStatusCode() !== 200) {
                return response()->json(['message' => __('could_not_logout')], 422);
            }

            $response = json_decode($response->getBody(), true);

            return response()->json($response);

        } catch (Exception $exception) {
            return response()->json(['error' => $exception->getMessage()], $exception->getCode());
        }
    }

    public function getPhonebooks(Request $request) {
        return $this->makeRequest('GET', 'phonebooks');
    }
    //Get phonebooks contacts from Web Socket and Http request.
    public function getPhonebookContacts($phonebookID, $isWsRequest = false) {
        $resource = 'phonebooks/' . $phonebookID . '/contacts';
        return $this->makeRequest('GET', $resource, null, $isWsRequest);
    }

    public function getVoicemail(Request $request) {
        try {
            if (!auth()->user()->extension_id) throw new Exception(__('you_do_not_have_a_device_assigned'), 422);

            $extension_id = auth()->user()->extension_id;
            $timezone = auth()->user()->searchSetting('general', 'timezone');
            $timezone = $timezone->value ?: 'UTC';
            $path = 'extensions/' . $extension_id . '/voicemail_records';

            $response = $this->makeRequest('GET', $path);

            if (
                !is_array($response) &&
                method_exists( $response, 'getStatusCode' ) &&
                $response->getStatusCode() !== 200
            ) {
                $errorMessage = __('could_not_get_voicemails');
                if ( $response->isClientError() && method_exists( $response, 'getOriginalContent' ) && is_array( $response->getOriginalContent() ) ) {
                    $errorMessage .= ": " . $response->getOriginalContent()['error'];
                }
                throw new Exception(
                    $errorMessage,
                    422
                );
            }

            $response = json_decode($response, true, 512, JSON_THROW_ON_ERROR);
            $voice_mails = (object) [
                'urgent' => [],
                'old' => [],
                'inbox' => []
            ];
            if (isset($response['data'])) {
                $data = $response['data'];
                $voice_mails = $this->parseVoicemailData($data, $voice_mails);
            }
            return response()->json([
                'voicemails' => $voice_mails
            ]);
        } catch (Exception | RequestException $e) {
            return response()->json([
                'message' => $e->getMessage(),
            ], 400);
        }
    }

	public function listenedVoicemail($folder, $filename) {
		$path = 'voicemail/' . auth()->user()->extension . '/' . $folder . '/' . $filename;
        return $this->makeRequest('LISTENED', $path);
	}

    public function deleteVoicemail(Request $request): JsonResponse
    {
        try {
            $folder = $request->input('folder');
            $voicemails = $request->input('voicemails');
            $updatedVoicemails = null;

            $folderMap = [
                'inbox' => 'INBOX',
                'old' => 'Old',
                'urgent' => 'Urgent'
            ];

            $folder = $folderMap[$folder] ?? $folder;

            foreach ($voicemails as $index => $value) {
                $msgId = $value['msg_id'];

                if ($index > 0) {
                    // Find the voicemail by msg_id in the updated list from the previous answer
                    $voicemail = $this->findVoicemailById($updatedVoicemails, $msgId);
                    if (!$voicemail) {
                        continue; // Skip if msg_id not found in updated list
                    }
                } else {
                    // For the first element, we directly use the filename of the request
                    $voicemail = $value;
                }

                $path = 'voicemail/' . auth()->user()->extension . '/' . $folder . '/' . $voicemail['filename'];

                $response = $this->makeRequest('DELETE', $path);

                if (
                    !is_array($response) &&
                    method_exists( $response, 'getStatusCode' ) &&
                    $response->getStatusCode() !== 200
                ) {
                    $errorMessage = '';
                    if ( $response->isClientError() && method_exists( $response, 'getOriginalContent' ) && is_array( $response->getOriginalContent() ) ) {
                        $errorMessage .= ": " . $response->getOriginalContent()['error'];
                    }
                    throw new Exception(
                        $errorMessage,
                        422
                    );
                }
                $response = json_decode($response, true, 512, JSON_THROW_ON_ERROR);
                if (isset($response['data'])) {
                    // Update the voicemail list with the list returned in the response
                    $updatedVoicemails = $response['data'];
                }
            }

            return response()->json([
                'message' => __('voicemail_actions_delete_success'),
                'voicemails' => $updatedVoicemails
            ]);
        }
        catch(Exception $e) {
            return response()->json([
                'message' => $e->getMessage()
            ], 400);
        }
    }

    /**
     * Function to find a specific voicemail by `msg_id`.
     */
    private function findVoicemailById($data, $msgId) {
        if (!is_null($data)) {
            foreach ($data as $value) {
                foreach ($value as $message) {
                    if ($message['msg_id'] === $msgId) {
                        return $message;
                    }
                }
            }
        }
        return null;
    }

    /**
     * @param Request $request
     *
     * @return JsonResponse|array
     */
    public function getCalls(Request $request): JsonResponse|array
    {
        try {

            $validated = $request->validate([
                'start_date' => 'sometimes|nullable|date',
                'end_date' => 'sometimes|nullable|date',
                'search' => 'sometimes|nullable|string',
                'disposition' => 'sometimes|nullable|in:NO_ANSWER,ANSWERED,BUSY,FAILED',
                'peer' => 'sometimes|nullable|string',
                'recordings' => 'sometimes|nullable|boolean',
                'offset' => 'required|integer|min:0',
                'row_count' => 'required|integer|min:1',
            ]);

            $user = auth()->user();
            $extension = $user->extension;

            if (is_null($extension)) {
                throw new UnexpectedValueException(
                    __('you_do_not_have_an_extension_assigned'),
                    422
                );
            }

            $timezone = $user->searchSetting('general', 'timezone')->value ?? 'UTC';

            $startDate = $validated['start_date'] ?? Carbon::now()->timezone($timezone)->subMonths(3);

            // if the start date is before the created_at date of the user, use the created_at date instead
            /*if (Carbon::parse($startDate)->isBefore($user->created_at)) {
                $startDate = Carbon::parse($user->created_at)->timestamp;
            }*/

            $endDate = $validated['end_date'] ?? Carbon::now()->timezone($timezone);
            $searchTerm = $request->has('search') && !is_null($request->input('search')) ? $validated['search'] : null;
            $offset = $validated['offset'];
            $rowCount = $validated['row_count'];
            $peer = $request->has('peer') && !is_null($request->input('peer')) ? $validated['peer'] : null;

            $startDate = Carbon::parse($startDate)->timezone($timezone)->timestamp;
            $endDate = Carbon::parse($endDate)->timezone($timezone)->timestamp;

            $extensions = !is_null($peer) ? $extension . ',' . $peer : $extension;
            $limit = $offset . ',' . $rowCount;
            $searchFields = 'disposition';

            $query = [
                'start_date' => $startDate,
                'end_date' => $endDate,
            ];
            if ($request->has('recordings') && !is_null($request->input('recordings'))) {
                $searchTerm = '/monitor/';
                $searchFields = 'recfile';
            } else {
                if ($request->has('disposition') && !is_null($request->input('disposition'))) {
                    $searchTerm = $validated['disposition'];
                } else {
                    if (!empty($searchTerm)) {
                        $searchFields = 'source,destination,clid';
                    }
                }
            }

            if (!is_null($peer)) {
                $query['call_history'] = 'yes';
            }

            $path  = 'cdr?extensions=' . $extensions . '&' . 'limit=' . $limit;
            if (!empty($searchTerm)) {
                $path .= '&search=' . $searchTerm . '&search_fields=' . $searchFields;
            }
            $path .= '&' . http_build_query($query);
            $response = $this->makeRequest('GET', $path, null, true);
            if (!is_array($response) && $response->getStatusCode() !== 200) {
                $errorMessage = __('could_not_get_calls');
                if ( $response->isClientError() && method_exists( $response, 'getOriginalContent' ) && is_array( $response->getOriginalContent() ) ) {
                    $errorMessage .= ": " . $response->getOriginalContent()['error'];
                }
                throw new UnexpectedValueException(
                    $errorMessage,
                    400
                );
            }

            $calls = [];

            if (isset($response['data']['result'])) {
                $data = $this->parseCallsData(
                    $response['data']['result'],
                    $timezone,
                    $request->has('recordings') && !is_null($request->input('recordings'))
                );
                $calls = collect($data->calls);
                // uniqueid is not unique (uniqueid is a unique identifier for the "Party A" channel and not for each CDR record)
                // $calls = $calls->unique('uniqueid');
                if ($request->has('recordings') && !is_null($request->input('recordings'))) {
                    $calls = $calls->unique('recfile');
                }
                $calls = $calls->sortByDesc('calldate')->values();
            }

            $responseData = [
                'calls' => $calls,
                'path'=> $path
            ];

            return response()->json($responseData);
        } catch (Exception | UnexpectedValueException $e) {
            return response()->json($e->getMessage(), 400);
        }
    }

    /**
     * Get the count of lost calls (calls with 'NO_ANSWER' disposition) for the authenticated user.
     *
     * This method retrieves the count of lost calls for the authenticated user by making an API
     * request to the CDR (Call Detail Records) endpoint. It takes into account the user's timezone
     * and calculates the time range to search for lost calls based on the user's last read timestamp
     * or the past two months.
     *
     * @return JsonResponse The response containing the last read timestamp and lost calls count.
     */
    /**
     * This code has been commented on, because after an investigation,
     * the "NO ANSWER" only occurs when the device is registered and the extension rings.
     * So, basically for this disposition to happen, the extension must be registered and, it must ring
     * Because of this, the management of "missed calls" is managed only within the life cycle of the app (frontend/client side).
     * REF: https://docs.asterisk.org/Configuration/Reporting/Call-Detail-Records-CDR/CDR-Specification/?h=busy#dispositions
     */
    /* public function getLostCallsCount()
    {
        try {
            $user = auth()->user();
            $extension = $user->extension;

            // Validate if the user has an assigned extension
            if (is_null($extension)) {
                throw new UnexpectedValueException(
                    __('you_do_not_have_an_extension_assigned'),
                    422
                );
            }

            // Determine the user's timezone, defaulting to 'UTC' if not set
            $timezone = $user->searchSetting('general', 'timezone')->value ?? 'UTC';

            // Set the default start date to two months ago
            $startDate = Carbon::now()
                ->timezone($timezone)
                ->subMonths(2)
                ->timestamp;

            // Check if the user has a last_calls_read timestamp
            if (!is_null($user->last_calls_read)) {
                $lastCallsRead = Carbon::parse($user->last_calls_read)->timezone($timezone);

                // Validate if the last_calls_read is within the last two months
                if ($lastCallsRead->greaterThanOrEqualTo(Carbon::now()->subMonths(2))) {
                    $startDate = $lastCallsRead->timestamp;
                }
            }

            // Set the end date to the current timestamp
            $endDate = Carbon::now()->timezone($timezone)->timestamp;

            // Define the query parameters for the API request
            $disposition = 'NO_ANSWER';
            $offset = 0;
            $rowCount = 19;
            $limit = $offset . ',' . $rowCount;
            $query = [
                'start_date' => $startDate,
                'end_date' => $endDate,
                'search' => $disposition,
                'search_fields' => 'disposition',
            ];

            // Build the API request path
            $path = 'cdr' . '?extensions=' . $extension . '&' . 'limit=' . $limit . '&' . http_build_query($query);

            // Make the API request
            $response = $this->makeRequest('GET', $path, null, true);

            // Validate the response
            if (!is_array($response) && $response->getStatusCode() !== 200) {
                $errorMessage = __('could_not_get_calls');
                if ($response->isClientError() && method_exists($response, 'getOriginalContent') && is_array($response->getOriginalContent())) {
                    $errorMessage .= ": " . $response->getOriginalContent()['error'];
                }
                throw new UnexpectedValueException($errorMessage, 400);
            }

            // Initialize the lost calls count
            $lostCalls = 0;
            if (isset($response['data']['result'])) {
                //$lostCalls = $response['data']['rows'];
                // we need to count the number of unique calls with the destination number based on the extension associated to the current user
                $lostCalls = collect($response['data']['result'])
                    ->where('destination', $extension)
                    ->count();
            }

            // Return the result as a JSON response
            return response()->json([
                'last_calls_read' => Carbon::createFromTimestamp($startDate)
                    ->timezone($timezone)
                    ->format('Y-m-d H:i:s'),
                'lost_calls' => $lostCalls
            ]);
        } catch (Exception | UnexpectedValueException $e) {
            // Handle any exceptions and return an error response
            return response()->json($e->getMessage(), 400);
        }
    } */

	public function formatCallerID ($str = '') {
        $name = '';
        $re = '/([A-Z])\w+/m';

        if (is_string($str) && !empty($str)) {
            preg_match_all($re, $str, $matches, PREG_SET_ORDER, 0);
            foreach ($matches as $key => $value) {
                $name.= $value[0];
                if (array_key_last($matches) !== $key) {
                    $name.= ' ';
                }
            }
        }
        return $name;
    }

    public function getExtensionFromCallerID ($clid = '') {
        try {
            if (!is_string($clid) || strtolower($clid) === 'unknown') {
                return null;
            }
            $re = '/\<(.*?)\>/m';
            preg_match_all($re, $clid, $matches, PREG_SET_ORDER, 0);
            if (empty($matches)) {
                return $clid;
            }
            return $matches[array_key_last($matches)][1];
        } catch (Exception $exception) {
            return null;
        }
    }

    public function findContactByPhone ($phone) {
        return Contact::where('user_id', auth()->id())
            ->whereHas('phones', function ($query) use ($phone) {
                $query->where('phone', $phone);
            })
            ->select('first_name', 'last_name')
            ->first();
    }

    /**
     * @param Request $request
     * This function requests the VitalPBX API to obtain the call history
     * and later obtain records with available recordings.
     * @return JsonResponse
     */

    public function getRecordings(GetCallsRequest $request): JsonResponse|array
    {
        return $this->getCalls($request);
    }

    public function parseCallsData ($data, $timezone, $getRecordings = false) {
        $user = auth()->user();
        $calls = [];
        // $lostCalls = 0;
        foreach ($data as $value) {
            if ($getRecordings && empty($value['recfile'])) {
                continue;
            }
            $number = $value['source'] !== $user->extension ? $value['source'] : $value['destination'];
            $call_info = (object) [
                /*'source' => $value['source'],
                'destination' => $value['destination'],*/
                'name' => $value['source'] !== $user->extension ? $this->formatCallerID($value['clid']) : $value['destination'],
                'original_name' =>  $value['source'] !== $user->extension ? $this->formatCallerID($value['clid']) : $value['destination'],
                'number' => $number ?? '',
                'calldate' => Carbon::createFromFormat('Y-m-d H:i:s', $value['calldate'], 'UTC')->setTimezone($timezone)->format('Y-m-d H:i:s'),
                'originator' => $value['destination'] === $user->extension ? 'remote' : 'local',
                'duration' => $value['duration'],
                'disposition' => $value['disposition'],
                'uniqueid' => $value['uniqueid'],
                'calltype' => $value['calltype'],
                'recfile' => $value['recfile'],
                'tenant' => $value['tenant'],
                'trunk' => $value['trunk'],
                'isUser' => false,
                'isContact' => false,
            ];

            // check if the number exists in the contacts
            $contact = $this->findContactByPhone($number);
            if (!is_null($contact)) {
                $call_info->full_name = ucfirst($contact->first_name) .' '. ucfirst($contact->last_name);
                $call_info->name = ucfirst($contact->first_name) .' '. ucfirst($contact->last_name);
                $call_info->avatar = $contact->getAvatarAttribute();
                $call_info->isContact = true;
            } else {
                // check if the number exists in the users
                $colleague = User::where('extension', $number ?? '')
                    ->where('tenant_id', auth()->user()->tenant_id)
                    ->first();
                if (!is_null($colleague)) {
                    $call_info->name = $colleague->name;
                    $call_info->avatar = $colleague->getAvatarAttribute();
                    $call_info->isUser = true;
                }
            }
            /**
             * This code has been commented on, because after an investigation,
             * the "NO ANSWER" only occurs when the device is registered and the extension rings.
             * So, basically for this disposition to happen, the extension must be registered and, it must ring
             * Because of this, the management of "missed calls" is managed only within the life cycle of the app (frontend/client side).
             * REF: https://docs.asterisk.org/Configuration/Reporting/Call-Detail-Records-CDR/CDR-Specification/?h=busy#dispositions
             */
            /* if (!is_null($user->last_calls_read)) {
                if ($value['destination'] === $user->extension && $value['disposition'] === 'NO ANSWER') {
                    if ($value['calldate'] > $user->last_calls_read) {
                        $lostCalls++;
                    }
                }
            } */

            $calls[] = $call_info;
        }
        return (object) [
            'calls' => $calls,
            // 'lost_calls' => $lostCalls
        ];
    }

    /**
     * @param mixed $data
     * @param object $voice_mails
     */
    public function parseVoicemailData(mixed $data, object $voice_mails): object
    {
        $timezone = auth()->user()->searchSetting('general', 'timezone');
        $timezone = $timezone->value ?: 'UTC';
        foreach ($data as $key => $value) {
            foreach ($value as $message) {
                $number = $this->getExtensionFromCallerID($message['clid']);
                $voice_mail = (object)[
                    'msg_id' => $message['msg_id'],
                    'name' => $this->formatCallerID($message['clid']),
                    'number' => $number ?? '',
                    'avatar' => null,
                    'date' => Carbon::createFromTimestamp($message['date'], 'UTC')->setTimezone($timezone)->format('Y-m-d H:i:s'),
                    'priority' => $message['priority'],
                    'vmorig' => $message['vmorig'],
                    'duration' => $message['duration'],
                    'recfile' => $message['recfile'],
                    'filename' => $message['filename'],
                    'isUser' => false,
                    'isContact' => false,
                ];

                $contact = $this->findContactByPhone($number);

                if ($contact) {
                    $voice_mail->name = $contact->first_name . ' ' . $contact->last_name;
                    $voice_mail->avatar = $contact->getAvatarAttribute();
                    $voice_mail->isContact = true;
                } else {
                    if (!is_null($number)) {
                        $user = User::where('extension', $number ?? '')
                            ->where('tenant_id', auth()->user()->tenant_id)
                            ->first();

                        if ($user) {
                            $voice_mail->name = $user->name;
                            $voice_mail->avatar = $user->getAvatarAttribute();
                            $voice_mail->isUser = true;
                        }
                    }
                }
                if (in_array($key, ['Urgent', 'urgent', 'URGENT'])) {
                    $voice_mails->urgent[] = $voice_mail;
                }
                if (in_array($key, ['Old', 'old', 'OLD'])) {
                    $voice_mails->old[] = $voice_mail;
                }
                if (in_array($key, ['Inbox', 'inbox', 'INBOX'])) {
                    $voice_mails->inbox[] = $voice_mail;
                }
            }
        }
        return $voice_mails;
    }

}
