<?php

namespace App\Http\Controllers;

use App\Http\Requests\StorePBXSettingsRequest;
use App\Utils\AppStatus;
use App\Utils\Environment;
use GuzzleHttp\Client;
use Illuminate\Support\Facades\Artisan;

class PbxSettingsController extends Controller
{

    private array $pbx_configs;

    public function __construct()
    {

        $installed = AppStatus::isInstalled();

        if (!$installed) {
            $this->middleware(['auth:api', 'role:super-admin'])->except('store');
        } else {
            $this->middleware(['auth:api', 'role:super-admin']);
        }

        $this->pbx_configs = [
            'host' => config('app.pbx.host'),
            'https_port' => config('app.pbx.https_port'),
            'pjsip_port' => config('app.pbx.pjsip_port'),
            'wss_port' => config('app.pbx.wss_port'),
            'api_port' => config('app.pbx.api_port'),
            'app_key' => config('app.pbx.app_key'),
        ];
    }

    public function show () {
        $pbx_configs = [
            'host' => $this->pbx_configs['host'],
            'https_port' => $this->pbx_configs['https_port'],
            'pjsip_port' => $this->pbx_configs['pjsip_port'],
            'wss_port' => $this->pbx_configs['wss_port'],
            'api_port' => $this->pbx_configs['api_port'],
        ];
        return response()->json($pbx_configs);
    }

    public function store (StorePBXSettingsRequest $request) {
        $validated = $request->validated();

        if ((empty($validated['app_key'])) && $this->checkKeyFile()) {
            $validated['app_key'] = $this->checkKeyFile();
        }

        if (!isset($validated['app_key'])) {
            return response()->json(['message' => __('app_key_is_required')], 422);
        }

        if ($this->checkConnection(
            $validated['host'],
            $validated['https_port'],
            $validated['app_key'],
        )) {

            $new_values = [
                'PBX_HOST' => $validated['host'],
                'PBX_HTTPS_PORT' => $validated['https_port'],
                'PJSIP_PORT' => $validated['pjsip_port'],
                'WSS_PORT' => $validated['wss_port'],
                'PBX_API_PORT' => $validated['api_port'],
                'PBX_APP_KEY' => $validated['app_key'],
            ];

            $updated = Environment::updateSettings($new_values);

            if (!$updated) {
                return response()->json(['message' => __('could_not_update')], 422);
            }

            Artisan::call('cache:clear');
            Artisan::call('config:clear');

            $this->pbx_configs['host'] = $validated['host'];
            $this->pbx_configs['https_port'] = $validated['https_port'];
            $this->pbx_configs['pjsip_port'] = $validated['pjsip_port'];
            $this->pbx_configs['wss_port'] = $validated['wss_port'];
            $this->pbx_configs['api_port'] = $validated['api_port'];
            $this->pbx_configs['app_key'] = $validated['app_key'];

            if ($this->checkKeyFile()) {
                $this->pbx_configs['app_key'] = '';
            }
            Artisan::call('websockets:restart');

            if (AppStatus::isInstalled()) {
                $eventLog = new PbxApiController();
                $user = auth()->user();
                $eventLog->saveEvents(
                    $user->tenant_id,
                    'PBX Settings',
                    'update',
                    $request->input(),
                    $user->user_name,
                    $user->id,
                    [
                    ],
                    $request->ip(),
                    $request->userAgent()
                );
            }

            return response()->json([
                'message' => __('updated'),
                'settings' => $this->pbx_configs
            ]);
        } else {
            return response()->json(
                ['message' => __('connection_could_not_be_established_please_check_your_settings')],
                422);
        }

        return response()->json(['message' => __('could_not_update')], 422);

    }
    /*This function check if exist the PBX Key File*/
    public function checkKeyFile()
    {
        $filePath = '/etc/vitalpbx/vitalpbx-maint.conf';

        if (is_file($filePath)) {
            return trim(file_get_contents($filePath));
        }

        return false;
    }

    public function checkConnection ($host, $port, $key) {
        $client = new Client();
        $protocol = 'https://';
        $version = 'v2';
        $endpoint = $protocol . $host . ':'. $port . '/api/' . $version . '/tenants';
        $connected = false;
        try {
            $request = $client->get($endpoint, [
                'headers' => [
                    'Accept'     => 'application/json',
                    'Content-Type' => 'application/json',
                    'Access-Control-Allow-Headers' => 'Authorization',
                    'app-key' => $key,
                ],
                'verify' => false
            ]);

            $connected = $request->getStatusCode() === 200;

        } catch (\Exception $exception) {
            // return $exception->getMessage();
        }
        return $connected;
    }

}
