<?php

namespace App\Http\Controllers;

use App\Http\Requests\RoleStoreRequest;
use App\Models\Role;
use Exception;
use Illuminate\Database\Eloquent\ModelNotFoundException;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Str;

class RolesController extends Controller
{
    public function index (): ?JsonResponse
    {
        try {
            $query = Role::with(['permissions' => function ($query) {
                $query->whereNotIn('slug', ['all-tenants', 'assigned-tenant']);
            }]);

            if (auth()->user()->hasRole('super-admin')) {
                $roles = $query->get();
            } else {
                $roles = $query->whereNotIn('slug', ['super-admin', 'admin'])->get();
            }

            return response()->json($roles);
        } catch (Exception $exception) {
            return response()->json([
                'message' => $exception->getMessage()
            ]);
        }
    }

    public function show (Request $request, Role $role): ?JsonResponse
    {
        try {
            $role->load('permissions');
            return response()->json($role);
        }
        catch (Exception $exception) {
            return response()->json([
                'message' => $exception->getMessage()
            ]);
        }
    }

    /**
     * Store a newly created resource in storage.
     *
     * @param RoleStoreRequest $request
     * @return JsonResponse
     */
    public function store (RoleStoreRequest $request): JsonResponse
    {
        try {
            $validated = $request->validated();

            $slug = Str::slug($validated['name'], '-');

            $role = new Role();
            $role->name = $validated['name'];
            $role->slug = $slug;
            $role->save();

            $role->permissions()->sync($validated['permissions']);

            $eventLog = new PbxApiController();

            $eventLog->saveEvents(
                auth()->user()->tenant_id,
                'Role Manager',
                'add',
                [
                    'name'=>$role->name,
                    "permissions" => $role->permissions()->pluck("slug")->toArray()
                ],
                auth()->user()->user_name,
                auth()->user()->id,
                [
                ],
                $request->ip(),
                $request->userAgent()
            );

            return response()->json([
                'message' => __('created_correctly'),
                'role' => $role->load('permissions')
            ]);
        } catch (ModelNotFoundException | Exception $exception) {
            return response()->json([
                'message' => $exception->getMessage()
            ], 422);
        }
    }

    /**
     * Update the specified resource in storage.
     *
     * @param RoleStoreRequest $request
     * @param \App\Models\Role $role
     * @return JsonResponse
     */
    public function update(RoleStoreRequest $request, Role $role): JsonResponse {
        try {
            $validated = $request->validated();

            $slug = Str::slug($validated['name'], '-');

            if ($slug === 'super-administrator') {
                $slug = 'super-admin';
            }
            if ($slug === 'administrator') {
                $slug = 'admin';
            }

            $role->name = $validated['name'];
            $role->slug = $slug;
            $role->save();

            if ($slug === 'super-admin') {
                $allTenantsPermission = DB::table('permissions')->where('slug', '=', 'all-tenants')->first();
                if (!is_null($allTenantsPermission)) {
                    $validated['permissions'][] = $allTenantsPermission->id;
                }
            }
            if ($slug === 'admin') {
                $assignedTenantPermission = DB::table('permissions')->where('slug', '=', 'assigned-tenant')->first();
                if (!is_null($assignedTenantPermission)) {
                    $validated['permissions'][] = $assignedTenantPermission->id;
                }
            }

            $role->permissions()->sync($validated['permissions']);

            $eventLog = new PbxApiController();
            $user_aut = auth()->user();
            $eventLog->saveEvents(
                $user_aut->tenant_id,
                'Role Manager',
                "update",
                [
                    'name'=>$role->name,
                    "permissions" => $role->permissions()->pluck("slug")->toArray()
                ],
                $user_aut->user_name,
                $user_aut->id,
                [
                ],
                $request->ip(),
                $request->userAgent()
            );

            return response()->json([
                'message' => __('updated'),
                'role' => $role->load('permissions')
            ]);
        } catch (ModelNotFoundException | Exception $exception) {
            return response()->json([
                'message' => $exception->getMessage()
            ], 422);
        }

    }

    /**
     * Remove the specified resource from storage.
     *
     * @param \App\Models\Role $role
     *
     * @return JsonResponse
     * @throws Exception
     */
    public function destroy (Role $role): JsonResponse {
        try {
            $protectedRoles = [1, 'super-admin', 'admin', 'user', 'agent'];

            if (in_array($role->id, $protectedRoles, true) || in_array($role->slug, $protectedRoles, true)) {
                throw new \Exception(__('it_is_not_allowed_to_delete_this_role'), 400);
            }

            $userRolesCount = DB::table('users_roles')
                                ->where('role_id', $role->id)
                                ->count();

            if ($userRolesCount > 0) {
                throw new \RuntimeException(__('there_are_users_who_have_assigned_this_role_please_edit_the_users_and_try_again'));
            }

            $permissions [] = $role->permissions()->pluck("name")->toArray();

            $role->delete();

            $eventLog = new PbxApiController();
            $user_aut = auth()->user();
            $client_ip = request()->ip();
            $device_info = request()->userAgent();
            $eventLog->saveEvents(
                $user_aut->tenant_id,
                'Role Manager',
                'delete',
                [
                    'name'=>$role->name,
                    "permissions" =>$permissions
                ],
                $user_aut->user_name,
                $user_aut->id,
                [
                ],
                $client_ip,
                $device_info
            );

            return response()->json([
                'message' => __('the_role_has_been_successfully_removed'),
                'role' => $role->id
            ]);
        } catch (Exception $exception) {
            return response()->json([
                'message' => $exception->getMessage(),
                'code' => $exception->getCode(),
                'file' => $exception->getFile(),
                'line' => $exception->getLine()
            ], 400);
        }
    }
}
