<?php

namespace App\Http\Controllers;

use Illuminate\Http\Request;
use Spatie\MediaLibrary\MediaCollections\Models\Media;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\StreamedResponse;

class ServePrivateStorage extends Controller
{

        /**
     * Validates the incoming request for serving private media files.
     *
     * @param Request $request The incoming request.
     * @param Media $media The media file to be served.
     * @param string $filename The expected filename of the media file.
     *
     * @return void If the request is valid, the function will not return anything.
     *             If the request is not valid, the function will abort with a 401 or 403 status code.
     *
     */
    private function validateRequest(Request $request, Media $media, string $filename): void
    {
        if (!$request->hasValidSignatureWhileIgnoring(['conversion', 'download'], false)) {
            abort(401);
        }

        if ($filename !== $media->file_name) {
            abort(403);
        }
    }

        /**
     * Serves private media files from the storage.
     *
     * This function handles the serving of private media files stored in the application's storage.
     * It validates the incoming request using a signature and checks the requested filename against the media file's name.
     * Depending on the presence of the 'download' parameter in the request, the function either returns a streamed response for downloading
     * or an inline response for viewing the media file in the browser.
     *
     * @param Request $request The incoming request.
     * @param Media $media The media file to be served.
     * @param string $filename The expected filename of the media file.
     *
     * @return StreamedResponse|Response The response containing the media file.
     *
     */
    public function show(Request $request, Media $media, $filename): StreamedResponse|Response
    {
        $this->validateRequest($request, $media, $filename);

        $conversion = null;
        if ($request->has('conversion')) {
            $conversion = $request->input('conversion', null);
        }

        if (!is_null($conversion)) {
            if ($media->hasGeneratedConversion($conversion)) {
                // Return the converted file
                return response()->file($media->getPath($conversion));
            } else {
                // Handle the case where the conversion doesn't exist
                abort(404, 'Requested conversion not available');
            }
        }

        return $request->has('download') ? $media->toResponse($request) : $media->toInlineResponse($request);
    }

}
