<?php

namespace App\Http\Controllers;

use App\Events\UserUpdate;
use App\Http\Requests\ChangePasswordRequest;
use App\Http\Requests\LoginRequest;
use App\Models\AppSetting;
use App\Models\Status;
use App\Models\User;
use Carbon\Carbon;
use Illuminate\Contracts\Auth\Guard;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\File;
use Illuminate\Support\Facades\Hash;
use Illuminate\Support\Facades\Http;
use Tymon\JWTAuth\Facades\JWTAuth;


class UserController extends Controller
{

    /**
     * Login username to be used by the controller.
     *
     * @var string
     */
    protected string $username;

    /**
     * Create a new AuthController instance.
     *
     * @return void
     */
    public function __construct()
    {
        $this->username = $this->findUsername();
    }

    /**
     * Get the token array structure.
     *
     * @param string $token
     *
     * @return JsonResponse
     * @throws \JsonException
     */
    protected function respondWithToken($token)
    {
        $user = JWTAuth::user($token);
        $device = null;
        if ($user->device_id) {
            $device = ( new PbxApiController() )->getDevice($user->device_id);
        }
        $pbxHost = config('app.pbx.host');

        $data = [
            'token' => $token,
            'user' => $user->load(
                'status',
                'groups:id',
                'settings:id,user_id,type,key,value',
                'roles:id,slug,name',
                'roles.permissions:id,name,slug',
                'agent_profile.profile:id,name,crm_url,website_tab_name,website_tab_url',
            ),
            'useful_data' => [
                'pbx_info' => [
                    'host' => $pbxHost,
                    'https_port' => config('app.pbx.https_port'),
                    'pjsip_port' => config('app.pbx.pjsip_port'),
                    'wss_port' => config('app.pbx.wss_port'),
                ],
                'socials' => $this->socialNetworks($user->tenant_id)
            ],
            'expires_in' => $this->guard()->factory()->getTTL() * 60
        ];

        if ($device && $device->getStatusCode() === 200) {
            $device = $device->getBody();
            $device = json_decode($device, false, 512, JSON_THROW_ON_ERROR);
            if ($device && $device->data && $device->data->secret) {
                $data['user']['device_secret'] = $device->data->secret;
            }
        }

        return response()->json($data);
    }

    /**
     * @param LoginRequest $request
     * @return JsonResponse
     */

    public function login(LoginRequest $request): JsonResponse
    {
        $validated = $request->validated();
        if ($token = $this->guard()->attempt(
            array(
                $this->username => $validated['login'],
                'password' => $validated['password']
            )
        )) {
            try {

                $user = JWTAuth::user($token);
                $eventLog = new PbxApiController();
                $eventLog->saveEvents(
                    $user->tenant_id,
                    'Core',
                    'login',
                    [
                    ],
                    $user->user_name,
                    $user->id,
                    [
                    ],
                    $request->ip(),
                    $request->userAgent()
                );

                return $this->respondWithToken($token);
            } catch (\JsonException $e) {
                //
            }
        }
        return response()->json(['error' => __('invalid_credentials')], 403);
    }

    /**
     * Get the login username to be used by the controller.
     *
     * @return string
     */
    public function findUsername() {
        $login = request()->input('login');

        $fieldType = filter_var($login, FILTER_VALIDATE_EMAIL) ? 'email' : 'user_name';

        request()->merge([$fieldType => $login]);

        return $fieldType;
    }

    /**
     * Get username property.
     *
     * @return string
     */
    public function username() {
        return $this->username;
    }

    /**
     * @param Request $request
     * @return JsonResponse
     */
    public function logout(Request $request)
    {
        if (is_null(auth()->user())) {
            return response()->json(['message' => "You are not logged"]);
        }
        $eventLog = new PbxApiController();
        $user = auth()->user();

        if ($user) {
            $eventLog->saveEvents(
                $user->tenant_id,
                'Core',
                'logout',
                [
                ],
                $user->user_name,
                $user->id,
                [
                ],
                $request->ip(),
                $request->userAgent()
            );
            $this->guard()->logout();
        }


        return response()->json(['message' => __('successfully_logged_out')]);

    }

    /**
     * Get the guard to be used during authentication.
     *
     * @return Guard
     */
    public function guard()
    {
        return Auth::guard();
    }

    public function me ($fromWsApi = false, $userId = null) {
        try {
            if (!is_null($userId)) {
                $user = User::where('id', $userId)->first();
            } else {
                $user = auth()->user();
            }

            $pbxHost = config('app.pbx.host');
            $pbxApiPort = config('app.pbx.api_port');
            $base = 'https://'. $pbxHost . ':'. $pbxApiPort;
            $apiVersion = 'v2';
            $sub = '/api/' . $apiVersion . '/';
            $resource = 'devices/' . $user->device_id;
            $device = null;
            // start request
            $response = Http::withOptions([
                'headers' => [
                    'Accept'     => 'application/json',
                    'Content-Type' => 'application/json',
                    'Access-Control-Allow-Headers' => 'Authorization',
                    'app-key' => config('app.pbx.app_key'),
                    'tenant' => $user->tenant_id,
                ],
                'verify' => false,
                'synchronous' => true
            ])->get($base . $sub . $resource);

            $ice_servers = AppSetting::getSetting('network', 'iceServers')->first();

            $userData = (object)  [
                'user' => User::where('id', $user->id)->with([
                        'status',
                        'groups:id',
                        'settings:id,user_id,type,key,value',
                        'roles:id,slug,name',
                        'roles.permissions:id,name,slug',
                        'agent_profile.profile:id,name,crm_url,website_tab_name,website_tab_url',
                    ]
                )->first(),
                'device' => null,
                'useful_data' => [
                    'pbx_info' => [
                        'host' => $pbxHost,
                        'https_port' => config('app.pbx.https_port'),
                        'pjsip_port' => config('app.pbx.pjsip_port'),
                        'wss_port' => config('app.pbx.wss_port')
                    ],
                    'ice_servers' => $ice_servers,
                    'socials' => $this->socialNetworks($user->tenant_id)
                ]
            ];

            if ($response->successful()) {
                $userData->user->device_secret = $response['data']['secret'];
                $userData->device = $response['data'];
            }
            if ($fromWsApi) {
                return $userData;
            }
            return response()->json($userData);
        } catch (Exception $e) {
            if ($fromWsApi) {
                return ['message' => $e->getMessage()];
            }
            return response()->json(['message' => $e->getMessage()]);
        }
    }
    //Get all groups.
    public function getGroups($fromWsApi = false)
    {
        $groups = auth()->user()->groups()->withCount('users')->get();

        return $fromWsApi ? $groups : response()->json($groups);
    }

    //Get all users from a group.
    public function getGroupUsers(Request $request, $id, $fromWsApi = false): JsonResponse|array
    {
        $group = auth()->user()->groups()->where('groups.id', '=', $id)->first();

        if (!$group) {
            return response()->json(['message' => 'not found'], 404);
        }

        // Handle pagination, sorting, and search defaults based on the request source
        $page = $request->input('page', 1);
        $perPage = $request->input('per_page', $fromWsApi ? 20 : 10);
        $searchTerm = $request->input('search', '');
        $sortBy = $request->input('sort_by', 'updated_at');
        $sortDesc = $request->input('sort_desc', 'desc');

        $usersQuery = User::select([
            'users.id',
            'users.name',
            'users.user_name',
            'users.extension',
            'users.email',
            'users.presence',
            'users.updated_at',
        ])
            ->with(['status' => function($query) {
                $query->select([
                    'status.id',
                    'status.name',
                    'status.slug',
                    'status.icon',
                    'status.color'
                ]);
            }])
            ->join('groups_users', 'users.id', '=', 'groups_users.user_id')
            ->leftJoin('user_status', 'users.id', '=', 'user_status.user_id')
            ->leftJoin('status', 'user_status.status_id', '=', 'status.id')
            ->where('groups_users.group_id', '=', $group->id);

        // Add search functionality
        if (!empty($searchTerm)) {
            $usersQuery->where(function ($query) use ($searchTerm) {
                $query->where('users.name', 'like', '%' . $searchTerm . '%')
                    ->orWhere('users.user_name', 'like', '%' . $searchTerm . '%')
                    ->orWhere('users.email', 'like', '%' . $searchTerm . '%')
                    ->orWhere('users.extension', 'like', '%' . $searchTerm . '%');
            });
        }

        // Paginate and order the users
        $users = $usersQuery
            ->orderBy($sortBy, $sortDesc)
            ->paginate($perPage, ['*'], 'page', $page);

        // Prepare the response structure
        $response = [
            'group' => $group,
            'users' => [
                'current_page' => $users->currentPage(),
                'data' => $users->items(),
                'last_page' => $users->lastPage(),
                'per_page' => $users->perPage(),
                'total' => $users->total(),
            ]
        ];

        return $fromWsApi ? $response : response()->json($response);
    }


    public function index (Request $request, $fromWsApi = false): JsonResponse|array
    {
        $searchTerm = $request->input('search', '');
        $filterBy = $request->input('filter_by', null);
        $page = $request->input('page', 1);
        $perPage = $request->input('per_page', 10);
        $sortBy = $request->input('sort_by', 'updated_at');
        $sortDesc = $request->input('sort_desc', 'desc');

        $groupsID = auth()->user()->groups()->get()->pluck('id')->toArray();
        $usersQuery = DB::table('users')
            ->distinct()
            ->select([
                'users.id',
                'users.name',
                'users.user_name',
                'users.extension',
                'users.email',
                'users.presence',
                'users.updated_at'
            ])
            ->join('groups_users', 'users.id', '=', 'groups_users.user_id')
            ->leftJoin('user_status', 'users.id', '=', 'user_status.user_id')
            ->leftJoin('status', 'user_status.status_id', '=', 'status.id')
            ->whereIn('groups_users.group_id', $groupsID);

        if (!empty($searchTerm)) {
            $usersQuery->where(function ($query) use ($searchTerm) {
                $query->where('users.name', 'like', '%' . $searchTerm . '%')
                    ->orWhere('users.user_name', 'like', '%' . $searchTerm . '%')
                    ->orWhere('users.email', 'like', '%' . $searchTerm . '%')
                    ->orWhere('users.extension', 'like', '%' . $searchTerm . '%');
            });
        }

        if (!is_null($filterBy)) {
            $usersQuery->whereNotNull('users.' . $filterBy);
        }

        $usersQuery->groupBy('users.id');

        $total = DB::table(DB::raw("({$usersQuery->toSql()}) as temp"))
            ->mergeBindings($usersQuery)
            ->count();
        // Paginate the results
        $users = $usersQuery
            ->orderBy('users.' . $sortBy, $sortDesc)
            ->paginate($perPage, ['*'], 'page', $page);

        $response = [
            'current_page' => $users->currentPage(),
            'data' => $users->items(),
            'last_page' => $users->lastPage(),
            'per_page' => $users->perPage(),
            'total' => $total,
        ];
        return $fromWsApi ? $response : response()->json($response);
    }

    public static function getUserByExtension ($extension, $tenant_id) {
        return DB::table('users')
            ->join('groups_users', 'groups_users.user_id', '=', 'users.id')
            ->select(
                'users.id',
                'users.name',
                'users.user_name',
                'users.extension',
                'users.device_name',
                'users.email',
                'users.presence',
            )
            // ->whereIn('groups_users.group_id', $groupsID)
            ->where('users.extension', '=', $extension)
            ->where('users.tenant_id', '=', $tenant_id)
            ->first();
    }

    public function markLostCallsAsRead () {
        try {
            auth()->user()->last_calls_read = new Carbon;
            auth()->user()->save();
            return response()->json([
                'message' => __('success')
            ]);
        } catch (\Exception $exception) {
            return response()->json([
                'message' => __('failed')
            ], 422);
        }
    }

    public function uploadAvatar (Request $request) {
        $user = auth()->user();

        $this->validate($request, [
            'image' => 'required|image|mimes:jpeg,png,jpg|max:2048',
        ]);

        try {

            if ($request->hasFile('image')) {
                $user->addMediaFromRequest('image')->toMediaCollection('avatar');
            } else {
                if ($user->hasMedia('avatar')) {
                    $user->clearMediaCollection('avatar');
                }
            }

            $user->refresh();
            try {
                broadcast(new UserUpdate($user, 'avatar', $user->avatar))->toOthers();
            } catch (\Exception $exception) {
                return response()->json([
                    'message' => __('broadcast_error'),
                    'info' => $exception->getMessage()
                ], 422);
            }

            return response()->json(
                [
                    'message' => __('success'),
                    'avatar' => $user->avatar
                ]
            );

        } catch (\Exception $exception) {
            return response()->json([
                'message' => $exception->getMessage()
            ], $exception->getCode());
        }
    }

    public function deleteAvatar (Request $request) {
        $user = auth()->user();
        if ($user->hasMedia('avatar')) {
            $user->clearMediaCollection('avatar');
        }
        return response()->json([
           'message' => __('avatar_deleted'),
           'avatar' => null
        ]);
    }

    public function getStatuses (): JsonResponse
    {
        return response()->json(Status::all());
    }

    public function updateStatus (Request $request) {
        $status_id = $request->input('status_id');
        $message = $request->input('message');

        auth()->user()->userStatus()->update([
            'message' => $message ?: null,
            'status_id' => $status_id
        ]);

        /**
         * If the issue fails, the process should continue.
         **/
        try {
            broadcast(new UserUpdate(auth()->user(),  'status', auth()->user()->status))->toOthers();
        } catch (\Exception $exception) {
            // nothing to do
        }

        return [
            'message' => __('status_updated'),
            'user' => auth()->user()->load('status')
        ];
    }

    public function setLocale (Request $request): JsonResponse
    {
        try {
            $this->validate($request, [
                'locale' => 'required|in:en,fr,es,pt,ko,de,ar,he,ru,it',
            ]);

            $user = auth()->user();
            $user->update([
                'locale' => $request->input('locale')
            ]);

            return response()->json([
                'message' => __('success'),
                'user' => auth()->user()
            ]);
        } catch (Exception $exception) {
            return response()->json(['message' => $exception->getMessage()]);
        }
    }

    public function socialNetworks ($tenant_id) {
        return DB::table('social_networks')->where('tenant_id', $tenant_id)->get();
    }

    public function getPauses () {
        try {
            $pauses = DB::table('users')
                ->join('user_agent_profile', 'users.id', '=', 'user_agent_profile.user_id')
                ->join('agent_profiles', 'user_agent_profile.agent_profile_id', '=', 'agent_profiles.id')
                ->join('agent_pauses', 'agent_profiles.id', '=', 'agent_pauses.agent_profile_id')
                ->select(
                    'user_agent_profile.user_id as user_id',
                    'agent_profiles.name as profile_name',
                    'agent_pauses.*'
                )
                ->groupBy('agent_pauses.id')
                ->where('users.id', '=', auth()->id())
                ->get();
            return response()->json(compact('pauses'));
        } catch (Exception $exception) {
            return response()->json($exception->getMessage(), 422);
        }
    }

    public function getPbxInfo () {
        return response()->json([
            'host' => config('app.pbx.host'),
            'https_port' => config('app.pbx.https_port'),
            'pjsip_port' => config('app.pbx.pjsip_port'),
            'wss_port' => config('app.pbx.wss_port')
        ]);
    }

    public function changePassword (ChangePasswordRequest $request) {
        $validated = $request->validated();

        auth()->user()->update([
            'password' => Hash::make($validated['password'])
        ]);

        return response()->json([
            'message' => __('successful_password_change')
        ]);
    }
}
